# -*- coding: UTF-8 -*-



import os

try:
    import cjson
    json_decoder = cjson.decode
except ImportError:
    import json
    json_decoder = json.loads

def get_bookmarks(bookmarks_file):
    # construct and configure the parser
    if not bookmarks_file:
        return []

    with open(bookmarks_file) as f:
        content = f.read()
        root = json_decoder(content)

    # make a dictionary of unique bookmarks
    bmap = {}

    def bmap_add(bmark, bmap):
        if bmark["id"] not in bmap:
            bmap[bmark["id"]] = bmark

    CONTAINER = "folder"
    UNWANTED_SCHEME = ("data", "place", "javascript")

    def is_container(ch):
        return ch["type"] == CONTAINER
    def is_bookmark(ch):
        return ch.get("url")
    def is_good(ch):
        return not ch["url"].split(":", 1)[0] in UNWANTED_SCHEME

    folders = []

    # add some folders
    folders.extend(root['roots']['bookmark_bar']['children'])
    folders.extend(root['roots']['other']['children'])

    for item in folders:
        if is_bookmark(item) and is_good(item):
            bmap_add(item, bmap)
        if is_container(item):
            folders.extend(item["children"])

    return list(bmap.values())

if __name__ == "__main__":
    fpath = os.path.expanduser("~/.config/chromium/Default/")
    print("Parsed # bookmarks:", len(list(get_bookmarks(fpath))))




"""
Implementation of unescaping and unquoting of the Exec= key in
the Desktop Entry Specification (As of March 2011, version 1.1-draft)
http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s06.html
http://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-1.1.html#exec-variables

The unescaping we are doing is only one way.. so we unescape according to the
rules, but we accept everything, if validly quoted or not.
"""

import shlex

# This is the "string" type encoding escapes
# this is unescaped before we process anything..
escape_table = {
    r'\s': ' ',
    r'\n': '\n',
    r'\t': '\t',
    r'\r': '\r',
    '\\\\': '\\',
}

# quoted are those chars that need a backslash in front
# (inside a double-quoted section, that is)
quoted = r""" " ` $ \ """.split()
quoted_table = {
    r'\"': '"',
    r'\`': '`',
    r'\$': '$',
    '\\\\': '\\',
}

'''
# reserved are those that need to be inside quotes
# note that all the quoted are also reserved, of course

We don't use these at all
reserved = r""" " ' \ > < ~ | & ; $ * ? # ( ) ` """.split()
reserved.extend([' ', '\t', '\n'])
'''

def two_part_unescaper(s, reptable):
    "Scan @s two characters at a time and replace using @reptable"
    if not s:
        return s
    def _inner():
        it = iter(zip(s, s[1:]))
        for cur, nex in it:
            key = cur+nex
            if key in reptable:
                yield reptable[key]
                try:
                    next(it)
                except StopIteration:
                    return
            else:
                yield cur
        yield s[-1]
    return ''.join(_inner())

def custom_shlex_split(s, comments=False, posix=True):
    """
    Wrapping shlex.split
    """
    if isinstance(s, str):
        is_unicode = True
        #s = s.encode("UTF-8")
    else:
        s = s.decode("UTF-8", "replace")
        is_unicode = False
    lex = shlex.shlex(s, posix=posix)
    lex.whitespace_split = True
    if not comments:
        lex.commenters = ''
    try:
        lex_output = list(lex)
    except ValueError:
        lex_output = [s]

    ## extra-unescape  ` and $ that are not handled by shlex
    quoted_shlex = {r'\`': '`', r'\$':'$'}
    lex_output[:] = [two_part_unescaper(x, quoted_shlex) for x in lex_output]
    if is_unicode:
        return lex_output
    else:
        return [x.encode("UTF-8") for x in lex_output]

def unescape(s):
    "Primary unescape of control sequences"
    return two_part_unescaper(s, escape_table)

def test_unescape():
    r"""
    >>> t = r'"This \\$ \\\\ \s\\\\"'
    >>> unescape(t)
    '"This \\$ \\\\  \\\\"'
    >>> unescape(r'\t\s\\\\')
    '\t \\\\'
    """
    pass

def parse_argv(instr):
    r"""
    Parse quoted @instr into an argv

    This is according to the spec
    >>> parse_argv('env "VAR=is good" ./program')
    ['env', 'VAR=is good', './program']
    >>> parse_argv('env "VAR=\\\\ \\$ @ x" ./program')
    ['env', 'VAR=\\ $ @ x', './program']
    >>> parse_argv('"\\$" "\\`"  "\\""')
    ['$', '`', '"']
    >>> parse_argv('/usr/bin/x-prog -q %F')
    ['/usr/bin/x-prog', '-q', '%F']
    >>> parse_argv('env LANG=en_US.UTF-8 freeciv-gtk2')
    ['env', 'LANG=en_US.UTF-8', 'freeciv-gtk2']
    >>> parse_argv('emacsclient -a "" -c %f')
    ['emacsclient', '-a', '', '-c', '%f']

    == Below this we need quirks mode ==

    The following style is common but not supported in spec
    >>> parse_argv('env VAR="is broken" ./program')
    ['env', 'VAR=is broken', './program']

    The following is just completely broken
    >>> parse_argv('./program unquoted\\\\argument')
    ['./program', 'unquoted\\argument']

    The following is just completely broken
    >>> parse_argv('./program No\\ Space')
    ['./program', 'No Space']

    The following is just insanely broken
    >>> parse_argv("'/opt'/now/'This is broken/'")
    ['/opt/now/This is broken/']

    This is broken
    #>>> parse_argv('\\$')
    #['$']
    #>>> parse_argv('\\$ \\`  \\"')
    #['$', '`', '"']

    Unmatched quote, normal mode (just testing that it does not raise)
    >>> parse_argv('"hi there')
    ['"hi there']

    Unmatched quote, quirks mode (just testing that it does not raise)
    >>> parse_argv('A\\\\BC "hi there')
    ['A\\\\BC "hi there']

    """
    return custom_shlex_split(instr)

def parse_unesc_argv(instr):
    r"""
    Parse quoted @instr into an argv after unescaping it

    >>> parse_unesc_argv(r'stuff "C:\\\\suck\\\\start.exe"')
    ['stuff', 'C:\\suck\\start.exe']

    == Below this we need quirks mode ==

    >>> parse_unesc_argv(r'stuff C:\\\\suck\\\\start.exe')
    ['stuff', 'C:\\suck\\start.exe']

    >>> parse_unesc_argv("'/usr'/bin/gnome-terminal -x gvim 'Insanely Broken'Yes")
    ['/usr/bin/gnome-terminal', '-x', 'gvim', 'Insanely BrokenYes']
    """
    return custom_shlex_split(unescape(instr))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



__kupfer_name__ = _("Search the Web")
__kupfer_sources__ = ("OpenSearchSource", )
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
        "SearchFor",
        "SearchWithEngine",
    )
__description__ = _("Search the web with OpenSearch search engines")
__version__ = "2020-04-19"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import locale
import os
import urllib.parse
import xml.etree.cElementTree as ElementTree

from kupfer.objects import Action, Source, Leaf
from kupfer.objects import TextLeaf
from kupfer import utils, config

from kupfer.plugin import firefox


def _noescape_urlencode(items):
    """Assemble an url param string from @items, without
    using any url encoding.
    """
    return "?" + "&".join("%s=%s" % (n, v) for n, v in items)


def _urlencode(word):
    """Urlencode a single string of bytes @word"""
    return urllib.parse.urlencode({"q": word})[2:]


def _do_search_engine(terms, search_url, encoding="UTF-8"):
    """Show an url searching for @search_url with @terms"""
    query_url = search_url.replace("{searchTerms}", _urlencode(terms))
    utils.show_url(query_url)


class SearchWithEngine (Action):
    """TextLeaf -> SearchWithEngine -> SearchEngine"""
    def __init__(self):
        Action.__init__(self, _("Search With..."))

    def activate(self, leaf, iobj):
        coding = iobj.object.get("InputEncoding")
        url = iobj.object["Url"]
        _do_search_engine(leaf.object, url, encoding=coding)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield SearchEngine

    def object_source(self, for_item=None):
        return OpenSearchSource()

    def get_description(self):
        return _("Search the web with OpenSearch search engines")

    def get_icon_name(self):
        return "edit-find"


class SearchFor (Action):
    """SearchEngine -> SearchFor -> TextLeaf

    This is the opposite action to SearchWithEngine
    """
    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj):
        coding = leaf.object.get("InputEncoding")
        url = leaf.object["Url"]
        terms = iobj.object
        _do_search_engine(terms, url, encoding=coding)

    def item_types(self):
        yield SearchEngine

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search the web with OpenSearch search engines")

    def get_icon_name(self):
        return "edit-find"


class SearchEngine (Leaf):
    def get_description(self):
        desc = self.object.get("Description")
        return desc if desc != str(self) else None

    def get_icon_name(self):
        return "text-html"


def coroutine(func):
    """Coroutine decorator: Start the coroutine"""
    def startcr(*ar, **kw):
        cr = func(*ar, **kw)
        next(cr)
        return cr
    return startcr


class OpenSearchParseError (Exception):
    pass


def gettagname(tag):
    return tag.rsplit("}", 1)[-1]


class OpenSearchSource (Source):
    def __init__(self):
        Source.__init__(self, _("Search Engines"))

    @coroutine
    def _parse_opensearch(self, target):
        """This is a coroutine to parse OpenSearch files"""
        vital_keys = set(["Url", "ShortName"])
        keys = set(["Description", "Url", "ShortName", "InputEncoding"])
        roots = ('OpenSearchDescription', 'SearchPlugin')

        def parse_etree(etree, name=None):
            if not gettagname(etree.getroot().tag) in roots:
                raise OpenSearchParseError("Search %s has wrong type" % name)
            search = {}
            for child in etree.getroot():
                tagname = gettagname(child.tag)
                if tagname not in keys:
                    continue
                # Only pick up Url tags with type="text/html"
                if tagname == "Url":
                    if (child.get("type") == "text/html" and
                            child.get("template")):
                        text = child.get("template")
                        params = {}
                        for ch in child:
                            if gettagname(ch.tag) == "Param":
                                params[ch.get("name")] = ch.get("value")
                        if params:
                            text += _noescape_urlencode(list(params.items()))
                    else:
                        continue
                else:
                    text = (child.text or "").strip()
                search[tagname] = text
            if not vital_keys.issubset(list(search.keys())):
                raise OpenSearchParseError("Search %s missing keys" % name)
            return search

        while True:
            try:
                path = (yield)
                etree = ElementTree.parse(path)
                target.send(parse_etree(etree, name=path))
            except Exception as exc:
                self.output_debug("%s: %s" % (type(exc).__name__, exc))

    def get_items(self):
        plugin_dirs = []

        # accept in kupfer data dirs
        plugin_dirs.extend(config.get_data_dirs("searchplugins"))

        # firefox in home directory
        ffx_home = firefox.get_firefox_home_file("searchplugins")
        if ffx_home and os.path.isdir(ffx_home):
            plugin_dirs.append(ffx_home)

        plugin_dirs.extend(config.get_data_dirs("searchplugins",
                                                package="firefox"))
        plugin_dirs.extend(config.get_data_dirs("searchplugins",
                                                package="iceweasel"))

        addon_dir = "/usr/lib/firefox-addons/searchplugins"
        cur_lang, _ignored = locale.getlocale(locale.LC_MESSAGES)
        suffixes = ["en-US"]
        if cur_lang:
            suffixes = [cur_lang.replace("_", "-"), cur_lang[:2]] + suffixes
        for suffix in suffixes:
            addon_lang_dir = os.path.join(addon_dir, suffix)
            if os.path.exists(addon_lang_dir):
                plugin_dirs.append(addon_lang_dir)
                break

        # debian iceweasel
        if os.path.isdir("/etc/iceweasel/searchplugins/common"):
            plugin_dirs.append("/etc/iceweasel/searchplugins/common")
        for suffix in suffixes:
            addon_dir = os.path.join("/etc/iceweasel/searchplugins/locale",
                                     suffix)
            if os.path.isdir(addon_dir):
                plugin_dirs.append(addon_dir)

        # try to find all versions of firefox
        for prefix in ('/usr/lib', '/usr/share'):
            for dirname in os.listdir(prefix):
                if dirname.startswith("firefox") or \
                        dirname.startswith("iceweasel"):
                    addon_dir = os.path.join(prefix, dirname,
                                             "searchplugins")
                    if os.path.isdir(addon_dir):
                        plugin_dirs.append(addon_dir)

                    addon_dir = os.path.join(prefix, dirname,
                                             "distribution", "searchplugins",
                                             "common")
                    if os.path.isdir(addon_dir):
                        plugin_dirs.append(addon_dir)

        self.output_debug("Found following searchplugins directories",
                          sep="\n", *plugin_dirs)

        @coroutine
        def collect(seq):
            """Collect items in list @seq"""
            while True:
                seq.append((yield))

        searches = []
        collector = collect(searches)
        parser = self._parse_opensearch(collector)
        # files are unique by filename to allow override
        visited_files = set()
        for pdir in plugin_dirs:
            try:
                for f in os.listdir(pdir):
                    if f in visited_files:
                        continue
                    fpath = os.path.join(pdir, f)
                    if os.path.isdir(fpath):
                        continue
                    parser.send(fpath)
                    visited_files.add(f)
            except EnvironmentError as exc:
                self.output_error(exc)

        for s in searches:
            yield SearchEngine(s, s["ShortName"])

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield SearchEngine

    def get_icon_name(self):
        return "applications-internet"



"""
It *should* be possible to support Tomboy and Gnote equally since
they support the same DBus protocol. This plugin takes this assumption.
"""

__kupfer_name__ = _("Notes")
__kupfer_sources__ = ("NotesSource", )
__kupfer_actions__ = (
        "AppendToNote",
        "AppendTextToNote",
        "CreateNote",
        "GetNoteSearchResults",
    )
__description__ = _("Gnote or Tomboy notes")
__version__ = "2017.2"
__author__ = ""

import os
import time
import xml.sax.saxutils

from gi.repository import GLib

import dbus
import xdg.BaseDirectory as base

from kupfer.objects import Action, Source, Leaf, TextLeaf, NotAvailableError, SourceLeaf
from kupfer.obj.apps import ApplicationSource
from kupfer.objects import TextSource
from kupfer import icons, plugin_support
from kupfer import pretty, textutils
from kupfer.weaklib import dbus_signal_connect_weakly


PROGRAM_IDS = ["gnote", "tomboy", "kzrnote"]
__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "notes_application",
        "label": _("Work with application"),
        "type": str,
        "value": "",
        "alternatives": ["",] + PROGRAM_IDS
    },
)

plugin_support.check_dbus_connection()

def unicode_strftime(fmt, time_tuple=None):
    return time.strftime(fmt, time_tuple)

## Tuples of  service name, object name, interface name
PROGRAM_SERIVCES = {
    "gnote": ("org.gnome.Gnote",
              "/org/gnome/Gnote/RemoteControl",
              "org.gnome.Gnote.RemoteControl"),
    "tomboy": ("org.gnome.Tomboy",
               "/org/gnome/Tomboy/RemoteControl",
               "org.gnome.Tomboy.RemoteControl"),
    "kzrnote": ("io.github.kupferlauncher.kzrnote",
                "/io/github/kupferlauncher/kzrnote",
                "io.github.kupferlauncher.kzrnote"),
}

def _get_notes_interface(activate=False):
    """Return the dbus proxy object for our Note Application.

    if @activate, we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()

    set_prog = __kupfer_settings__["notes_application"]
    programs = (set_prog, ) if set_prog else PROGRAM_IDS

    for program in programs:
        service_name, obj_name, iface_name = PROGRAM_SERIVCES[program]
        if activate:
            _ignore, status = bus.start_service_by_name(service_name)
        else:
            if not bus.name_has_owner(service_name):
                continue

        try:
            searchobj = bus.get_object(service_name, obj_name)
        except dbus.DBusException as e:
            pretty.print_error(__name__, e)
            return
        return dbus.Interface(searchobj, iface_name)

def _get_notes_interactive():
    """
    Return the dbus proxy object, activate if necessary,
    raise an OperationError if not available.
    """
    obj = _get_notes_interface(activate=True)
    if obj is None:
        raise NotAvailableError(__kupfer_settings__["notes_application"])
    return obj

def reply_noop(*args):
    pass

class RetryDbusCalls(pretty.OutputMixin):
    """
    A d-bus interface wrapper for a proxy object; will retry a method
    call if it fails (a limited number of times)

    The method call must be async (with reply_handler and error_handler)
    """
    def __init__(self, proxy_object, retries=10):
        self.__obj = proxy_object
        self.__retries = retries

    @property
    def proxy_obj(self):
        """
        Return the inner proxy object. You can call methods synchronously on it.
        """
        return self.__obj

    def __getattr__(self, name):
        x = 0
        def proxy_method(*args, error_handler=None, **kwargs):
            def make_call():
                getattr(self.__obj, name)(*args, error_handler=error_handler_, **kwargs)
            def error_handler_(exc):
                nonlocal x
                x += 1
                if (x > self.__retries or
                    exc.get_dbus_name() != "org.freedesktop.DBus.Error.UnknownMethod"):
                    return error_handler(exc)
                self.output_debug("retrying", name, "because of", exc)
                GLib.timeout_add(25 * x, make_call)
            return make_call()
        return proxy_method

def make_error_handler(ctx):
    def error_handler(exc):
        pretty.print_debug(__name__, exc)
        ctx.register_late_error(NotAvailableError(__kupfer_settings__["notes_application"]))
    return error_handler

class Open (Action):
    action_accelerator = "o"
    def __init__(self):
        Action.__init__(self, _("Open"))
    def wants_context(self):
        return True
    def activate(self, leaf, ctx):
        noteuri = leaf.object
        notes = RetryDbusCalls(_get_notes_interactive())
        notes.DisplayNote(noteuri,
                          reply_handler=reply_noop,
                          error_handler=make_error_handler(ctx))
    def get_description(self):
        return _("Open with notes application")
    def get_gicon(self):
        app_icon = icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)
        return icons.ComposedIcon(self.get_icon_name(), app_icon)

class AppendToNote (Action):
    def __init__(self):
        Action.__init__(self, _("Append to Note..."))

    def wants_context(self):
        return True
    def activate(self, leaf, iobj, ctx):
        notes = RetryDbusCalls(_get_notes_interactive())
        noteuri = iobj.object
        text = leaf.object

        def reply_note_plain_text(contents):
            if not contents.endswith("\n"):
                contents += "\n"
            contents += text
            if not text.endswith("\n"):
                contents += "\n"
            notes.SetNoteContents(noteuri, contents,
                                  reply_handler=reply_noop,
                                  error_handler=make_error_handler(ctx))

        def reply_note_xml(xmlcontents):
            # NOTE: We search and replace in the XML here
            endtag = "</note-content>"
            xmltext = xml.sax.saxutils.escape(text)
            xmlcontents = xmlcontents.replace(endtag, "\n%s%s" % (xmltext, endtag))
            notes.SetNoteCompleteXml(noteuri, xmlcontents,
                                     reply_handler=reply_noop,
                                     error_handler=make_error_handler(ctx))

        if __kupfer_settings__["notes_application"] == "kzrnote":
            notes.GetNoteContents(noteuri,
                                  reply_handler=reply_note_plain_text,
                                  error_handler=make_error_handler(ctx))
        else:
            notes.GetNoteCompleteXml(noteuri,
                                     reply_handler=reply_note_xml,
                                     error_handler=make_error_handler(ctx))

    def item_types(self):
        yield TextLeaf
    def requires_object(self):
        return True
    def object_types(self):
        yield Note
    def object_source(self, for_item=None):
        return NotesSource()
    def get_description(self):
        return _("Add text to existing note")
    def get_icon_name(self):
        return "list-add"

class AppendTextToNote (AppendToNote):
    def __init__(self):
        Action.__init__(self, _("Append..."))
    def wants_context(self):
        return True
    def activate(self, leaf, iobj, ctx):
        return super().activate(iobj, leaf, ctx)

    def item_types(self):
        return super().object_types()
    def requires_object(self):
        return True
    def object_types(self):
        return super().item_types()
    def object_source(self, for_item=None):
        return TextSource()

def _prepare_note_text(text):
    ## split the text into a title + newline + rest of the text
    ## if we only get the title, put in two helpful newlines
    title, body = textutils.extract_title_body(text)
    if body.lstrip():
        return "%s\n%s" % (title, body)
    else:
        return "%s\n\n" % (title,)

class CreateNote (Action):
    def __init__(self):
        Action.__init__(self, _("Create Note"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        notes = RetryDbusCalls(_get_notes_interactive())
        text = _prepare_note_text(leaf.object)

        def _created_note(noteuri):
            nonlocal notes
            notes = notes.proxy_obj
            # FIXME: For Gnote we have to call DisplayNote
            # else we can't change its contents
            notes.DisplayNote(noteuri)
            notes.SetNoteContents(noteuri, text)

        notes.CreateNote(reply_handler=_created_note,
                         error_handler=make_error_handler(ctx))


    def item_types(self):
        yield TextLeaf
    def get_description(self):
        return _("Create a new note from this text")
    def get_icon_name(self):
        return "document-new"

class GetNoteSearchResults (Action):
    def __init__(self):
        Action.__init__(self, _("Get Note Search Results..."))

    def is_factory(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        query = leaf.object.lower()
        notes = RetryDbusCalls(_get_notes_interactive())

        def search_reply(noteuris):
            ctx.register_late_result(NoteSearchSource(query, noteuris))

        notes.SearchNotes(query, False,
                          reply_handler=search_reply,
                          error_handler=make_error_handler(ctx))

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Show search results for this query")

class NoteSearchSource (Source):
    def __init__(self, query, noteuris):
        self.query = query.lower()
        Source.__init__(self, _("Notes"))
        self.noteuris = noteuris

    def get_items(self):
        notes = _get_notes_interactive()
        for noteuri in self.noteuris:
            title = notes.GetNoteTitle(noteuri)
            date = notes.GetNoteChangeDate(noteuri)
            yield Note(noteuri, title, date)

    def repr_key(self):
        return self.query

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)

    def provides(self):
        yield Note

class Note (Leaf):
    """The Note Leaf's represented object is the Note URI"""
    def __init__(self, obj, name, date):
        self.changedate = date
        Leaf.__init__(self, obj, name)
    def get_actions(self):
        yield Open()
    def repr_key(self):
        # the Note URI is unique&persistent for each note
        return self.object
    def get_description(self):
        today_date = time.localtime()[:3]
        yest_date = time.localtime(time.time() - 3600*24)[:3]
        change_time = time.localtime(self.changedate)

        if today_date == change_time[:3]:
            time_str = _("today, %s") % unicode_strftime("%X", change_time)
        elif yest_date == change_time[:3]:
            time_str = _("yesterday, %s") % unicode_strftime("%X", change_time)
        else:
            time_str = unicode_strftime("%c", change_time)
        # TRANS: Note description, %s is last changed time in locale format
        return _("Last updated %s") % time_str
    def get_icon_name(self):
        return "text-x-generic"

class ClassProperty (property):
    """Subclass property to make classmethod properties possible"""
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

class NotesSource (ApplicationSource):
    def __init__(self):
        Source.__init__(self, _("Notes"))
        self._notes = []

    def initialize(self):
        """Set up filesystem monitors to catch changes"""
        # We monitor all directories that exist of a couple of candidates
        dirs = []
        for program in PROGRAM_IDS:
            notedatapaths = (os.path.join(base.xdg_data_home, program),
                    os.path.expanduser("~/.%s" % program))
            dirs.extend(notedatapaths)
        self.monitor_token = self.monitor_directories(*dirs)

        set_prog = __kupfer_settings__["notes_application"]
        if set_prog in PROGRAM_SERIVCES:
            bus_name = PROGRAM_SERIVCES[set_prog][0]
            bus = dbus.SessionBus()
            dbus_signal_connect_weakly(bus, "NameOwnerChanged", self._name_owner_changed,
                                       dbus_interface="org.freedesktop.DBus",
                                       arg0=bus_name)

    def _name_owner_changed(self, name, old, new):
        if new:
            self.mark_for_update()

    def _update_cache(self, notes):
        try:
            noteuris = notes.ListAllNotes()
        except dbus.DBusException as e:
            self.output_error("%s: %s" % (type(e).__name__, e))
            return

        templates = notes.GetAllNotesWithTag("system:template")

        self._notes = []
        for noteuri in noteuris:
            if noteuri in templates:
                continue
            title = notes.GetNoteTitle(noteuri)
            date = notes.GetNoteChangeDate(noteuri)
            self._notes.append((noteuri, title, date))

    def get_items(self):
        notes = _get_notes_interface()
        if notes:
            self._update_cache(notes)
        for noteuri, title, date in self._notes:
            yield Note(noteuri, title, date=date)

    def provides(self):
        yield Note

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, PROGRAM_IDS)

    def get_icon_name(self):
        return "gnote"

    @ClassProperty
    @classmethod
    def appleaf_content_id(cls):
        return __kupfer_settings__["notes_application"] or PROGRAM_IDS




# encoding: utf-8

from kupfer import icons
from kupfer import pretty
from kupfer import utils
from kupfer import datatools
from kupfer import puid

from kupfer.obj.base import Leaf, Action, Source, InvalidDataError
from kupfer.obj.objects import Perform, RunnableLeaf, TextLeaf

class TimedPerform (Perform):
    """A timed (delayed) version of Run (Perform) """
    action_accelerator = None
    def __init__(self):
        Action.__init__(self, _("Run after Delay..."))

    def activate(self, leaf, iobj, ctx):
        from kupfer import scheduler
        # make a timer that will fire when Kupfer exits
        interval = utils.parse_time_interval(iobj.object)
        pretty.print_debug(__name__, "Run %s in %s seconds" % (leaf, interval))
        timer = scheduler.Timer(True)
        args = (ctx,) if leaf.wants_context() else ()
        timer.set(interval, leaf.run, *args)

    def requires_object(self):
        return True
    def object_types(self):
        yield TextLeaf

    def valid_object(self, iobj, for_item=None):
        interval = utils.parse_time_interval(iobj.object)
        return interval > 0

    def get_description(self):
        return _("Perform command after a specified time interval")

class ComposedLeaf (RunnableLeaf):
    serializable = 1
    def __init__(self, obj, action, iobj=None):
        object_ = (obj, action, iobj)
        # A slight hack: We remove trailing ellipsis and whitespace
        format = lambda o: str(o).strip(".… ")
        name = " → ".join([format(o) for o in object_ if o is not None])
        RunnableLeaf.__init__(self, object_, name)

    def __getstate__(self):
        state = dict(vars(self))
        state["object"] = [puid.get_unique_id(o) for o in self.object]
        return state

    def __setstate__(self, state):
        vars(self).update(state)
        objid, actid, iobjid = state["object"]
        obj = puid.resolve_unique_id(objid)
        act = puid.resolve_action_id(actid, obj)
        iobj = puid.resolve_unique_id(iobjid)
        if (not obj or not act) or (iobj is None) != (iobjid is None):
            raise InvalidDataError("Parts of %s not restored" % str(self))
        self.object[:] = [obj, act, iobj]

    def get_actions(self):
        yield Perform()
        yield TimedPerform()

    def repr_key(self):
        return self

    def wants_context(self):
        return True

    def run(self, ctx):
        obj, action, iobj = self.object
        return ctx.delegated_run(obj, action, iobj)

    def get_gicon(self):
        obj, action, iobj = self.object
        return icons.ComposedIcon(obj.get_icon(), action.get_icon())

class _MultipleLeafContentSource (Source):
    def __init__(self, leaf):
        Source.__init__(self, str(leaf))
        self.leaf = leaf
    def get_items(self):
        return self.leaf.object

class MultipleLeaf (Leaf):
    """
    A Leaf for the direct representation of many leaves. It is not
    a container or "source", it *is* the many leaves itself.

    The represented object is a sequence of Leaves
    """
    serializable = 1
    def __init__(self, obj, name=_("Multiple Objects")):
        # modifying the list of objects is strictly forbidden
        robj = list(datatools.UniqueIterator(obj))
        Leaf.__init__(self, robj, name)

    def get_multiple_leaf_representation(self):
        return self.object

    def __getstate__(self):
        state = dict(vars(self))
        state["object"] = [puid.get_unique_id(o) for o in self.object]
        return state

    def __setstate__(self, state):
        vars(self).update(state)
        objects = []
        for id_ in state["object"]:
            obj = puid.resolve_unique_id(id_)
            if obj is None:
                raise InvalidDataError("%s could not be restored!" % (id_, ))
            objects.append(obj)
        self.object[:] = objects

    def has_content(self):
        return True

    def content_source(self, alternate=False):
        return _MultipleLeafContentSource(self)

    def get_description(self):
        n = len(self.object)
        return ngettext("%s object", "%s objects", n) % (n, )
    def get_icon_name(self):
        return "kupfer-object-multiple"



__kupfer_name__ = _("Favorites")
__kupfer_sources__ = ("FavoritesSource", )
__kupfer_actions__ = ("AddFavorite", "RemoveFavorite", )
__description__ = _("Mark commonly used items and store objects for later use")
__version__ = "2009-12-30"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import weakref

from kupfer.objects import Leaf, Source, Action
from kupfer import puid

# NOTE: core import
from kupfer.core import learn


class FavoritesSource (Source):
    """Keep a list of Leaves that the User may add and remove from"""
    instance = None
    def __init__(self):
        Source.__init__(self, _("Favorites"))
        ## these are default favorites for new users
        self.references = [
                '<kupfer.plugin.core.contents.Help>',
                '<kupfer.plugin.core.contents.Preferences>',
            ]

    def config_save(self):
        references = [puid.get_unique_id(F) for F in self.favorites]
        return {"favorites": references, "version": self.version}

    def config_save_name(self):
        return __name__

    def config_restore(self, state):
        self.references = state["favorites"]

    def _lookup_item(self, id_):
        itm = puid.resolve_unique_id(id_, excluding=self)
        if itm is None:
            return None
        return itm

    def _valid_item(self,  itm):
        if hasattr(itm, "is_valid") and not itm.is_valid():
            return False
        return True

    def _find_item(self, id_):
        itm = self._lookup_item(id_)
        if itm is None or not self._valid_item(itm):
            return None
        if puid.is_reference(id_):
            self.reference_table[id_] = itm
        else:
            self.persist_table[id_] = itm
        return itm

    def initialize(self):
        FavoritesSource.instance = self
        self.favorites = []
        self.persist_table = {}
        self.reference_table = weakref.WeakValueDictionary()
        self.mark_for_update()

    def _update_items(self):
        self.favorites = []
        self.mark_for_update()
        for id_ in self.references:
            if id_ in self.persist_table:
                self.favorites.append(self.persist_table[id_])
                continue
            if id_ in self.reference_table:
                self.favorites.append(self.reference_table[id_])
                continue
            itm = self._find_item(id_)
            if itm is None:
                self.output_debug("MISSING:", id_)
            else:
                self.favorites.append(itm)

    @classmethod
    def add(cls, itm):
        cls.instance._add(itm)

    def _add(self, itm):
        if self._has_item(itm):
            self._remove(itm)
        learn.add_favorite(itm)
        self.favorites.append(itm)
        self.references.append(puid.get_unique_id(itm))
        self.mark_for_update()

    @classmethod
    def has_item(cls, itm):
        return cls.instance._has_item(itm)

    def _has_item(self, itm):
        return itm in set(self.favorites)

    @classmethod
    def remove(cls, itm):
        if cls.has_item(itm):
            cls.instance._remove(itm)

    def _remove(self, itm):
        learn.remove_favorite(itm)
        self.favorites.remove(itm)
        id_ = puid.get_unique_id(itm)
        if id_ in self.references:
            self.references.remove(id_)
        else:
            for key, val in self.persist_table.items():
                if val == itm:
                    self.references.remove(key)
                    self.persist_table.pop(key)
                    break
        self.mark_for_update()

    def get_items(self):
        self._update_items()
        for fav in self.favorites:
            learn.add_favorite(fav)
        return reversed(self.favorites)

    def get_description(self):
        return _('Shelf of "Favorite" items')

    def get_icon_name(self):
        return "emblem-favorite"

    def provides(self):
        # returning nothing means it provides anything
        return ()

class AddFavorite (Action):
    # Rank down, since it applies everywhere
    rank_adjust = -15
    def __init__(self):
        Action.__init__(self, _("Add to Favorites"))
    def activate(self, leaf):
        FavoritesSource.add(leaf)
    def item_types(self):
        return (Leaf, )
    def valid_for_item(self, item):
        return not FavoritesSource.has_item(item)
    def get_description(self):
        return _("Add item to favorites shelf")
    def get_icon_name(self):
        return "list-add"

class RemoveFavorite (Action):
    rank_adjust = -15
    def __init__(self):
        Action.__init__(self, _("Remove from Favorites"))
    def activate(self, leaf):
        FavoritesSource.remove(leaf)
    def item_types(self):
        return (Leaf, )
    def valid_for_item(self, item):
        return FavoritesSource.has_item(item)
    def get_description(self):
        return _("Remove item from favorites shelf")
    def get_icon_name(self):
        return "list-remove"



import itertools
import os
from os import path as os_path
import locale
import signal
import sys

from gi.repository import GLib, Gtk, Gio, Gdk

import functools


from kupfer import pretty
from kupfer import kupferstring
from kupfer import desktop_launch
from kupfer import launch
from kupfer import desktop_parse
from kupfer import terminal

from kupfer.desktop_launch import SpawnError


def get_dirlist(folder, depth=0, include=None, exclude=None):
    """
    Return a list of absolute paths in folder
    include, exclude: a function returning a boolean
    def include(filename):
        return ShouldInclude
    """
    from os import walk
    paths = []
    def include_file(file):
        return (not include or include(file)) and (not exclude or not exclude(file))
        
    for dirname, dirnames, fnames in walk(folder):
        # skip deep directories
        head, dp = dirname, 0
        while not os_path.samefile(head, folder):
            head, tail = os_path.split(head)
            dp += 1
        if dp > depth:
            del dirnames[:]
            continue
        
        excl_dir = []
        for dir in dirnames:
            if not include_file(dir):
                excl_dir.append(dir)
                continue
            abspath = os_path.join(dirname, dir)
            paths.append(abspath)
        
        for file in fnames:
            if not include_file(file):
                continue
            abspath = os_path.join(dirname, file)
            paths.append(abspath)

        for dir in reversed(excl_dir):
            dirnames.remove(dir)

    return paths

def locale_sort(seq, key=str):
    """Return @seq of objects with @key function as a list sorted
    in locale lexical order

    >>> locale.setlocale(locale.LC_ALL, "C")
    'C'
    >>> locale_sort("abcABC")
    ['A', 'B', 'C', 'a', 'b', 'c']

    >>> locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
    'en_US.UTF-8'
    >>> locale_sort("abcABC")
    ['a', 'A', 'b', 'B', 'c', 'C']
    """
    locale_cmp = lambda s, o: locale.strcoll(key(s), key(o))
    seq = seq if isinstance(seq, list) else list(seq)
    seq.sort(key=functools.cmp_to_key(locale_cmp))
    return seq

def _argv_to_locale(argv):
    "encode unicode strings in @argv according to the locale encoding"
    return [kupferstring.tolocale(A) if isinstance(A, str) else A
            for A in argv]

class AsyncCommand (pretty.OutputMixin):
    """
    Run a command asynchronously (using the GLib mainloop)

    call @finish_callback when command terminates, or
    when command is killed after @timeout_s seconds, whichever
    comes first.

    If @timeout_s is None, no timeout is used

    If stdin is a byte string, it is supplied on the command's stdin.

    If env is None, command will inherit the parent's environment.

    finish_callback -> (AsyncCommand, stdout_output, stderr_output)

    Attributes:
    self.exit_status  Set after process exited
    self.finished     bool

    """
    # the maximum input (bytes) we'll read in one shot (one io_callback)
    max_input_buf = 512 * 1024

    def __init__(self, argv, finish_callback, timeout_s, stdin=None, env=None):
        self.stdout = []
        self.stderr = []
        self.stdin = []
        self.timeout = False
        self.killed = False
        self.finished = False
        self.finish_callback = finish_callback

        # FIXME: No byte string support
        self.output_debug("AsyncCommand:", argv)

        flags = (GLib.SPAWN_SEARCH_PATH | GLib.SPAWN_DO_NOT_REAP_CHILD)
        kwargs = {}
        if env is not None:
            kwargs['envp'] = env
        pid, stdin_fd, stdout_fd, stderr_fd = \
             GLib.spawn_async(argv, standard_output=True, standard_input=True,
                              standard_error=True, flags=flags, **kwargs)

        if stdin:
            self.stdin[:] = self._split_string(stdin, self.max_input_buf)
            in_io_flags = GLib.IO_OUT | GLib.IO_ERR | GLib.IO_HUP | GLib.IO_NVAL
            GLib.io_add_watch(stdin_fd, in_io_flags, self._in_io_callback,
                              self.stdin)
        else:
            os.close(stdin_fd)

        io_flags = GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP | GLib.IO_NVAL
        GLib.io_add_watch(stdout_fd, io_flags, self._io_callback, self.stdout)
        GLib.io_add_watch(stderr_fd, io_flags, self._io_callback, self.stderr)
        self.pid = pid
        GLib.child_watch_add(pid, self._child_callback)
        if timeout_s is not None:
            GLib.timeout_add_seconds(timeout_s, self._timeout_callback)

    def _split_string(self, s, length):
        """Split @s in pieces of @length"""
        L = []
        for i in range(0, len(s)//length + 1):
            L.append(s[i*length:(i+1)*length])
        return L

    def _io_callback(self, sourcefd, condition, databuf):
        if condition & GLib.IO_IN:
            databuf.append(os.read(sourcefd, self.max_input_buf))
            return True
        return False

    def _in_io_callback(self, sourcefd, condition, databuf):
        """write to child's stdin"""
        if condition & GLib.IO_OUT:
            if not databuf:
                os.close(sourcefd)
                return False
            s = databuf.pop(0)
            written = os.write(sourcefd, s)
            if written < len(s):
                databuf.insert(0, s[written:])
            return True
        return False

    def _child_callback(self, pid, condition):
        # @condition is the &status field of waitpid(2) (C library)
        self.exit_status = os.WEXITSTATUS(condition)
        self.finished = True
        self.finish_callback(self, b"".join(self.stdout), b"".join(self.stderr))

    def _timeout_callback(self):
        "send term signal on timeout"
        if not self.finished:
            self.timeout = True
            os.kill(self.pid, signal.SIGTERM)
            GLib.timeout_add_seconds(2, self._kill_callback)

    def _kill_callback(self):
        "Last resort, send kill signal"
        if not self.finished:
            self.killed = True
            os.kill(self.pid, signal.SIGKILL)


def spawn_terminal(workdir=None, screen=None):
    " Raises SpawnError "
    term = terminal.get_configured_terminal()
    notify = term["startup_notify"]
    app_id = term["desktopid"]
    argv = term["argv"]
    desktop_launch.spawn_app_id(app_id, argv, workdir, notify, screen)

def spawn_in_terminal(argv, workdir=None):
    " Raises SpawnError "
    term = terminal.get_configured_terminal()
    notify = term["startup_notify"]
    _argv = list(term["argv"])
    if term["exearg"]:
        _argv.append(term["exearg"])
    _argv.extend(argv)
    desktop_launch.spawn_app_id(term["desktopid"], _argv , workdir, notify)

def spawn_async_notify_as(app_id, argv):
    """
    Spawn argument list @argv and startup-notify as
    if application @app_id is starting (if possible)

    raises SpawnError
    """
    desktop_launch.spawn_app_id(app_id, argv , None, True)

def spawn_async(argv, in_dir="."):
    """
    Silently spawn @argv in the background

    Returns False on failure
    """
    try:
        return spawn_async_raise(argv, in_dir)
    except SpawnError as exc:
        pretty.print_debug(__name__, "spawn_async", argv, exc)
        return False

def spawn_async_raise(argv, workdir="."):
    """
    A version of spawn_async that raises on error.

    raises SpawnError
    """
    # FIXME: How to support locale strings?
    argv
    #argv = _argv_to_locale(argv)
    pretty.print_debug(__name__, "spawn_async", argv, workdir)
    try:
        return GLib.spawn_async (argv, working_directory=workdir,
                flags=GLib.SPAWN_SEARCH_PATH)
    except GLib.GError as exc:
        raise SpawnError(exc.message)

def argv_for_commandline(cli):
    return desktop_parse.parse_argv(cli)

def launch_commandline(cli, name=None, in_terminal=False):
    " Raises SpawnError "
    argv = desktop_parse.parse_argv(cli)
    pretty.print_error(__name__, "Launch commandline is deprecated ")
    pretty.print_debug(__name__, "Launch commandline (in_terminal=", in_terminal, "):", argv, sep="")
    if in_terminal:
        return spawn_in_terminal(argv)
    return spawn_async(argv)

def launch_app(app_info, files=(), uris=(), paths=()):
    " Raises SpawnError "

    # With files we should use activate=False
    return launch.launch_application(app_info, files, uris, paths,
            activate=False)

def show_path(path):
    """Open local @path with default viewer"""
    # Implemented using Gtk.show_uri
    gfile = Gio.File.new_for_path(path)
    if not gfile:
        return
    url = gfile.get_uri()
    show_url(url)

def show_url(url):
    """Open any @url with default viewer"""
    try:
        pretty.print_debug(__name__, "show_url", url)
        return Gtk.show_uri(Gdk.Screen.get_default(), url, Gtk.get_current_event_time())
    except GLib.GError as exc:
        pretty.print_error(__name__, "Gtk.show_uri:", exc)

def _on_child_exit(pid, condition, user_data):
    # @condition is the &status field of waitpid(2) (C library)
    argv, respawn = user_data
    if respawn:
        is_signal = os.WIFSIGNALED(condition)
        if is_signal and respawn:
            def callback(*args):
                spawn_child(*args)
                return False
            GLib.timeout_add_seconds(10, callback, argv, respawn)

def _try_register_pr_pdeathsig():
    """
    Register PR_SET_PDEATHSIG (linux-only) for the calling process
    which is a signal delivered when its parent dies.

    This should ensure child processes die with the parent.
    """
    PR_SET_PDEATHSIG=1
    SIGHUP=1
    if sys.platform != 'linux2':
        return
    try:
        import ctypes
    except ImportError:
        return
    try:
        libc = ctypes.CDLL("libc.so.6")
        libc.prctl(PR_SET_PDEATHSIG, SIGHUP)
    except (AttributeError, OSError):
        pass

def spawn_child(argv, respawn=True, display=None):
    """
    Spawn argv in the mainloop and keeping it as a child process
    (it will be made sure to exit with the parent).

    @respawn: If True, respawn if child dies abnormally

    raises utils.SpawnError
    returns pid
    """
    flags = (GLib.SPAWN_SEARCH_PATH | GLib.SPAWN_DO_NOT_REAP_CHILD)
    envp = []
    if display:
        # environment is passed as a sequence of strings
        envd = os.environ.copy()
        envd['DISPLAY'] = display
        envp[:] = ['='.join((k,v)) for k,v in list(envd.items())]

    try:
        pid, stdin_fd, stdout_fd, stderr_fd = \
            GLib.spawn_async(argv, envp,flags=flags,
                             child_setup=_try_register_pr_pdeathsig,
                             )
    except GLib.GError as exc:
        raise SpawnError(str(exc))
    if pid:
        GLib.child_watch_add(pid, _on_child_exit, (argv, respawn))
    return pid

def start_plugin_helper(name, respawn, display=None):
    """
    @respawn: If True, respawn if child dies abnormally

    raises SpawnError
    """
    argv = [sys.executable]
    argv.extend(sys.argv)
    argv.append('--exec-helper=%s' % name)
    pretty.print_debug(__name__, "Spawning", argv)
    return spawn_child(argv, respawn, display=display)

def show_help_url(url):
    """
    Try at length to display a startup notification for the help browser.

    Return False if there is no handler for the help URL
    """
    ## Check that the system help viewer is Yelp,
    ## and if it is, launch its startup notification.
    scheme = Gio.File.new_for_uri(url).get_uri_scheme()
    default = Gio.app_info_get_default_for_uri_scheme(scheme)
    help_viewer_id = "yelp.desktop"
    if not default:
        return False
    try:
        yelp = Gio.DesktopAppInfo.new(help_viewer_id)
    except (TypeError, RuntimeError):
        return show_url(url)
    cmd_path = lookup_exec_path(default.get_executable())
    yelp_path = lookup_exec_path(yelp.get_executable())
    if cmd_path and yelp_path and os.path.samefile(cmd_path, yelp_path):
        try:
            spawn_async_notify_as(help_viewer_id, [cmd_path, url])
            return True
        except SpawnError:
            pass
    return show_url(url)

def lookup_exec_path(exename):
    "Return path for @exename in $PATH or None"
    PATH = os.environ.get("PATH") or os.defpath
    for execdir in PATH.split(os.pathsep):
        exepath = os.path.join(execdir, exename)
        if os.access(exepath, os.R_OK|os.X_OK) and os.path.isfile(exepath):
            return exepath

def is_directory_writable(dpath):
    """If directory path @dpath is a valid destination to write new files?
    """
    if not os_path.isdir(dpath):
        return False
    return os.access(dpath, os.R_OK | os.W_OK | os.X_OK)

def get_destpath_in_directory(directory, filename, extension=None):
    """Find a good destpath for a file named @filename in path @directory
    Try naming the file as filename first, before trying numbered versions
    if the previous already exist.

    If @extension, it is used as the extension. Else the filename is split and
    the last extension is used
    """
    # find a nonexisting destname
    ctr = itertools.count(1)
    basename = filename + (extension or "")
    destpath = os_path.join(directory, basename)
    while True:
        if not os_path.exists(destpath):
            break
        if extension:
            root, ext = filename, extension
        else:
            root, ext = os_path.splitext(filename)
        basename = "%s-%s%s" % (root, next(ctr), ext)
        destpath = os_path.join(directory, basename)
    return destpath

def get_destfile_in_directory(directory, filename, extension=None):
    """Find a good destination for a file named @filename in path @directory.

    Like get_destpath_in_directory, but returns an open file object, opened
    atomically to avoid race conditions.

    Return (fileobj, filepath)
    """
    # retry if it fails
    for retry in range(3):
        destpath = get_destpath_in_directory(directory, filename, extension)
        try:
            fd = os.open(destpath, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
        except OSError as exc:
            pretty.print_error(__name__, exc)
        else:
            return (os.fdopen(fd, "wb"), destpath)
    return (None, None)

def get_safe_tempfile():
    """Return (fileobj, filepath) pointing to an open temporary file"""
    import tempfile
    fd, path = tempfile.mkstemp()
    return (os.fdopen(fd, "wb"), path)

def get_display_path_for_bytestring(filepath):
    """Return a unicode path for display for bytestring @filepath

    Will use glib's filename decoding functions, and will
    format nicely (denote home by ~/ etc)
    """
    desc = GLib.filename_display_name(filepath)
    homedir = os.path.expanduser("~/")
    if desc.startswith(homedir) and homedir != desc:
        desc = desc.replace(homedir, "~/", 1)
    return desc

def parse_time_interval(tstr):
    """
    Parse a time interval in @tstr, return whole number of seconds

    >>> parse_time_interval("2")
    2
    >>> parse_time_interval("1h 2m 5s")
    3725
    >>> parse_time_interval("2 min")
    120
    """
    weights = {
        "s": 1, "sec": 1,
        "m": 60, "min": 60,
        "h": 3600, "hours": 3600,
    }
    try:
        return int(tstr)
    except ValueError:
        pass

    total = 0
    amount = 0
    # Split the string in runs of digits and runs of characters
    for isdigit, group in itertools.groupby(tstr, lambda k: k.isdigit()):
        part = "".join(group).strip()
        if not part:
            continue
        if isdigit:
            amount = int(part)
        else:
            total += amount * weights.get(part.lower(), 0)
            amount = 0
    return total


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from gi.repository import Gio

from kupfer.objects import FileLeaf

def register_async_file_result(ctx, filepath):
    """
    Register that @filepath may appear soon
    @ctx: The action's execution context token
    """
    return AsyncFileResult(ctx, filepath)

class AsyncFileResult (object):
    """Expect a given file path to be created, and when (probably) done,
    post the file as a late result.
    """
    def __init__(self, ctx, filepath):
        self.ctx = ctx
        gfile = Gio.File.new_for_path(filepath)
        self.monitor = gfile.monitor_file(Gio.FileMonitorFlags.NONE)
        self.callback_id = self.monitor.connect("changed", self.changed)

    def changed(self, monitor, gfile1, gfile2, event):
        if event == Gio.FileMonitorEvent.CHANGES_DONE_HINT:
            self.ctx.register_late_result(FileLeaf(gfile1.get_path()))
            self.monitor.disconnect(self.callback_id)
            self.monitor = None





# Copyright (C) 2009  Ulrik Sverdrup <ulrik.sverdrup@gmail.com>
#               2008  Christian Hergert <chris@dronelabs.com>
#               2007  Chris Halse Rogers, DR Colkitt
#                     David Siegel, James Walker
#                     Jason Smith, Miguel de Icaza
#                     Rick Harding, Thomsen Anders
#                     Volker Braun, Jonathon Anderson 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import division

"""
This module provides relevance matching and formatting of related strings
based on the relevance.  It originates in Gnome-Do.

 * Python port by Christian Hergert

 * Module updated by Ulrik Sverdrup to clean up and dramatically speed up
   the code, by using more pythonic constructs as well as doing less work.

Compatibility: Python 2.4 and later, including Python 3
"""



# This module is compatible with both Python 2 and Python 3;
# we need the iterator form of range for either version, stored in range()
try:
    range = xrange
except NameError:
    pass

def formatCommonSubstrings(s, query, format_clean=None, format_match=None):
    """
    Creates a new string highlighting matching substrings.

    Returns: a formatted string

    >>> formatCommonSubstrings('hi there dude', 'hidude',
    ...                        format_match=lambda m: "<b>%s</b>" % m)
    '<b>hi</b> there <b>dude</b>'

    >>> formatCommonSubstrings('parallelism', 'lsm', format_match=str.upper)
    'paralleLiSM'
    """
    format_clean = format_clean or (lambda x: x)
    format_match = format_match or (lambda x: x)
    format = lambda x: x and format_clean(x)

    if not query:
        return format(s)

    ls = s.lower()

    # find overall range of match
    first, last = _findBestMatch(ls, query)

    if first == -1:
        return format(s)

    # find longest perfect match, put in slc
    for slc in range(len(query), 0, -1):
        if query[:slc] == ls[first:first+slc]:
            break
    nextkey = query[slc:]

    head = s[:first]
    match = s[first: first+slc]
    matchtail = s[first+slc: last]
    tail = s[last:]

    # we use s[0:0], which is "" or u""
    return s[0:0].join((
            format(head),
            format_match(match),
            formatCommonSubstrings(matchtail, nextkey,
                                   format_clean, format_match),
            format(tail),
            ))

def score_single(s, query):
    """
    s: text body to score
    query: A single character

    This is a single character approximation to `score`.

    >>> round(score_single('terminal', 't'), 6)
    0.973125
    >>> round(score_single('terminal', 'e'), 6)
    0.903125
    >>> round(score_single('terminal', 'a'), 6)
    0.903125
    >>> round(score_single('t', 't'), 6)
    0.995
    """

    ls = s.lower()

    # Find the shortest possible substring that matches the query
    # and get the ration of their lengths for a base score
    first = ls.find(query)
    if first == -1:
        return .0
    score = 0.9 + .025 / len(s)

    if first == 0:
        score += 0.07
    return score

def score(s, query):
    """
    A relevancy score for the string ranging from 0 to 1

    @s: a string to be scored
    @query: a string query to score against

    `s' is treated case-insensitively while `query' is interpreted literally,
    including case and whitespace.

    Returns: a float between 0 and 1

    >>> round(score('terminal', 'trml'), 6)
    0.735099
    >>> round(score('terminal', 'term'), 6)
    0.992303
    >>> print(score('terminal', 'try'))
    0.0
    >>> print(score('terminal', ''))
    1.0
    >>> round(score('terminal', 't'), 6)
    0.98949
    >>> round(score('terminal', 'e'), 6)
    0.918438
    """
    if not query:
        return 1.0

    ls = s.lower()

    # Find the shortest possible substring that matches the query
    # and get the ration of their lengths for a base score
    first, last = _findBestMatch(ls, query)
    if first == -1:
        return .0

    score = len(query) / (last - first)

    # Now we weight by string length so shorter strings are better
    score *= .7 + len(query) / len(s) * .3

    # Bonus points if the characters start words
    good = 0
    bad = 1
    firstCount = 0
    for i in range(first, last-1):
        if ls[i] in " -.([_":
            if ls[i + 1] in query:
                firstCount += 1
            else:
                bad += 1
    
    # A first character match counts extra
    if query[0] == ls[0]:
        firstCount += 2
        
    # The longer the acronym, the better it scores
    good += firstCount * firstCount * 4
    
    # Better yet if the match itself started there
    if first == 0:
        good += 2

    # Super duper bonus if it is a perfect match
    if query == ls:
        good += last * 2 + 4

    score = (score + 3 * good / (good + bad)) / 4

    # This fix makes sure that perfect matches always rank higher
    # than split matches.  Perfect matches get the .9 - 1.0 range
    # everything else lower
    
    if last - first == len(query):
        score = .9 + .1 * score
    else:
        score = .9 * score
    
    return score

def _findBestMatch(s, query):
    """
    Finds the shortest substring of @s that contains all characters of query
    in order.

    @s: a string to search
    @query: a string query to search for

    Returns: a two-item tuple containing the start and end indicies of
             the match.  No match returns (-1,-1).

    >>> _findBestMatch('terminal', 'trml')
    (0, 8)
    >>> _findBestMatch('total told', 'tl')
    (2, 5)
    >>> _findBestMatch('terminal', 'yl')
    (-1, -1)
    """
    bestMatch = -1, -1
    
    # Find the last instance of the last character of the query
    # since we never need to search beyond that
    lastChar = s.rfind(query[-1])
    
    # No instance of the character?
    if lastChar == -1:
        return bestMatch

    # Loop through each instance of the first character in query
    index = s.find(query[0])

    queryLength = len(query)
    lastIndex = lastChar - len(query) + 1
    while 0 <= index <= lastIndex:
        # See if we can fit the whole query in the tail
        # We know the first char matches, so we dont check it.
        cur = index + 1
        qcur = 1
        while qcur < queryLength:
            # find where in the string the next query character is
            # if not found, we are done
            cur = s.find(query[qcur], cur, lastChar + 1)
            if cur == -1:
                return bestMatch
            cur += 1
            qcur += 1

        # take match if it is shorter
        # if perfect match, we are done
        if bestMatch[0] == -1 or (cur - index) < (bestMatch[1] - bestMatch[0]):
            bestMatch = (index, cur)
            if cur - index == queryLength:
                break

        index = s.find(query[0], index + 1)

    return bestMatch

if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Module for confiugration and misc things
"""

import xdg.BaseDirectory as base
import os

PACKAGE_NAME="kupfer"

class ResourceLookupError (Exception):
    pass

def has_capability(cap):
    return not bool(os.getenv("KUPFER_NO_%s" % cap, False))

def get_kupfer_env(name, default=""):
    return os.getenv("KUPFER_%s" % name, default)

def get_cache_home():
    """
    Directory where cache files should be put
    Guaranteed to exist
    """
    cache_home = base.xdg_cache_home or os.path.expanduser("~/.cache")
    cache_dir = os.path.join(cache_home, PACKAGE_NAME)
    if not os.path.exists(cache_dir):
        try:
            os.makedirs(cache_dir, mode=0o700)
        except OSError as e:
            print(e)
            return None
    return cache_dir

def get_cache_file(path=()):
    cache_home = base.xdg_cache_home or os.path.expanduser("~/.cache")
    cache_dir = os.path.join(cache_home, *path)
    if not os.path.exists(cache_dir):
        return None
    return cache_dir

def get_data_file(filename, package=PACKAGE_NAME):
    """
    Return path to @filename if it exists
    anywhere in the data paths, else raise ResourceLookupError.
    """
    data_paths = []
    try:
        from . import version_subst
    except ImportError:
        first_datadir = "./data"
    else:
        first_datadir = os.path.join(version_subst.DATADIR, package)

    data_paths.append(first_datadir)
    for data_path in base.load_data_paths(package):
        if not data_path in data_paths:
            data_paths.append(data_path)

    for direc in data_paths:
        file_path = os.path.join(direc, filename)
        if os.path.exists(file_path):
            return file_path
    if package == PACKAGE_NAME:
        raise ResourceLookupError("Resource %s not found" % filename)
    else:
        raise ResourceLookupError("Resource %s in package %s not found" %
            (filename, package))

def save_data_file(filename):
    """
    Return filename in the XDG data home directory, where the
    directory is guaranteed to exist
    """
    direc = base.save_data_path(PACKAGE_NAME)
    if not direc:
        return None
    filepath = os.path.join(direc, filename)
    return filepath

def get_data_home():
    """
    Directory where data is to be saved
    Guaranteed to exist
    """
    return base.save_data_path(PACKAGE_NAME)

def get_data_dirs(name="", package=PACKAGE_NAME):
    """
    Iterate over all data dirs of @name that exist
    """
    return base.load_data_paths(os.path.join(package, name))

def get_config_file(filename, package=PACKAGE_NAME):
    """
    Return path to @package/@filename if it exists anywhere in the config
    paths, else return None
    """
    return base.load_first_config(package, filename)

def get_config_files(filename):
    """
    Iterator to @filename in all
    config paths, with most important (takes precendence)
    files first
    """
    return base.load_config_paths(PACKAGE_NAME, filename) or ()

def save_config_file(filename):
    """
    Return filename in the XDG data home directory, where the
    directory is guaranteed to exist
    """
    direc = base.save_config_path(PACKAGE_NAME)
    if not direc:
        return None
    filepath = os.path.join(direc, filename)
    return filepath



# -*- encoding: UTF-8 -*-

import locale
from unicodedata import normalize, category

def _folditems():
    _folding_table = {
        # general non-decomposing characters
        # FIXME: This is not complete
        "ł" : "l",
        "œ" : "oe",
        "ð" : "d",
        "þ" : "th",
        # germano-scandinavic canonical transliterations
        "ü" : "ue",
        "å" : "aa",
        "ä" : "ae",
        "æ" : "ae",
        "ö" : "oe",
        "ø" : "oe",
    }

    for c, rep in list(_folding_table.items()):
        yield (ord(c.upper()), rep.upper())
        yield (ord(c), rep)
    yield ord("ß"), "ss"
    yield ord("ẞ"), "SS"

folding_table = dict(_folditems())

def tounicode(utf8str):
    """Return `unicode` from UTF-8 encoded @utf8str
    This is to use the same error handling etc everywhere
    """
    if isinstance(utf8str, str):
        return utf8str
    return utf8str.decode("UTF-8", "replace") if utf8str is not None else ""

def toutf8(ustr):
    """Return UTF-8 `str` from unicode @ustr
    This is to use the same error handling etc everywhere
    if ustr is `str`, just return it
    """
    if isinstance(ustr, bytes):
        return ustr
    return ustr.encode("UTF-8")

def fromlocale(lstr):
    """Return a unicode string from locale bytestring @lstr"""
    assert isinstance(lstr, bytes)
    enc = locale.getpreferredencoding(do_setlocale=False)
    return lstr.decode(enc, "replace")

def tolocale(ustr):
    """Return a locale-encoded bytestring from unicode @ustr"""
    assert isinstance(ustr, str)
    enc = locale.getpreferredencoding(do_setlocale=False)
    return ustr.encode(enc)


def tofolded(ustr):
    """Fold @ustr

    Return a unicode string where composed characters are replaced by
    their base, and extended latin characters are replaced by
    similar basic latin characters.

    >>> tofolded(u"Wyłącz")
    'Wylacz'
    >>> tofolded(u"naïveté")
    'naivete'
    >>> tofolded(u"Groß")
    'Gross'

    Characters from other scripts are not transliterated.

    >>> print(tofolded(u"Ἑλλάς"))
    Ελλας
    """
    srcstr = normalize("NFKD", ustr.translate(folding_table))
    return "".join([c for c in srcstr if category(c) != 'Mn'])

if __name__ == '__main__':
    import doctest
    doctest.testmod()




def _get_leaf_members(leaf):
    """
    Return an iterator to members of @leaf, if it is a multiple leaf
    """
    try:
        return leaf.get_multiple_leaf_representation()
    except AttributeError:
        return (leaf, )

def action_valid_for_item(action, leaf):
    return all(action.valid_for_item(L) for L in _get_leaf_members(leaf))

def actions_for_item(leaf, sourcecontroller):
    if leaf is None:
        return []
    actions = None
    for L in _get_leaf_members(leaf):
        l_actions = set(L.get_actions())
        l_actions.update(sourcecontroller.get_actions_for_leaf(L))
        if actions is None:
            actions = l_actions
        else:
            actions.intersection_update(l_actions)
    return actions

def iobject_source_for_action(action, for_item):
    """
    Return (src, use_catalog)

    where
    src: object source or None,
    use_catalog: True to use catalog in addition.
    """
    for leaf in _get_leaf_members(for_item):
        return action.object_source(leaf), action.object_source_and_catalog(leaf)

def iobjects_valid_for_action(action, for_item):
    """
    Return a filtering *function* that will let through
    those leaves that are good iobjects for @action and @for_item.
    """
    def valid_object(leaf, for_item):
        _valid_object = action.valid_object
        for L in _get_leaf_members(leaf):
            for I in _get_leaf_members(for_item):
                if not _valid_object(L, for_item=I):
                    return False
        return True

    types = tuple(action.object_types())
    def type_obj_check(iobjs):
        for i in iobjs:
            if (isinstance(i, types) and valid_object(i, for_item=for_item)):
                yield i
    def type_check(itms):
        for i in itms:
            if isinstance(i, types):
                yield i

    if hasattr(action, "valid_object"):
        return type_obj_check
    else:
        return type_check




__kupfer_name__ = _("Trash")
__kupfer_actions__ = ("MoveToTrash", "EmptyTrash")
__kupfer_sources__ = ("TrashSource", )
__description__ = _("Access trash contents")
__version__ = "2017.2"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

from gi.repository import Gio, GLib

from kupfer.objects import Leaf, Action, Source, SourceLeaf, FileLeaf
from kupfer.objects import OperationError
from kupfer.obj.fileactions import Open
from kupfer import utils, icons, pretty


TRASH_URI = 'trash://'

class MoveToTrash (Action):
    # this should never be default
    rank_adjust = -10
    def __init__(self):
        Action.__init__(self, _("Move to Trash"))

    def activate(self, leaf):
        gfile = leaf.get_gfile()
        try:
            gfile.trash()
        except GLib.Error as exc:
            raise OperationError(exc.message)

    def valid_for_item(self, item):
        gfile = item.get_gfile()
        if not gfile.query_exists(None):
            return False
        info = gfile.query_info(Gio.FILE_ATTRIBUTE_ACCESS_CAN_TRASH,
                Gio.FileQueryInfoFlags.NONE, None)
        return info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_TRASH)
    def get_description(self):
        return _("Move this file to trash")
    def get_icon_name(self):
        return "user-trash-full"
    def item_types(self):
        yield FileLeaf


class RestoreTrashedFile (Action):
    def __init__(self):
        Action.__init__(self, _("Restore"))

    def has_result(self):
        return True

    def activate(self, leaf):
        orig_path = leaf.get_orig_path()
        if not orig_path:
            return
        orig_gfile = Gio.File.new_for_path(orig_path)
        cur_gfile = leaf.get_gfile()
        if orig_gfile.query_exists():
            raise OperationError("Target file exists at %s" % orig_gfile.get_path())
        pretty.print_debug(__name__, "Move %s to %s" % (cur_gfile, orig_gfile))
        ret = cur_gfile.move(orig_gfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None)
        pretty.print_debug(__name__, "Move ret=%s" % (ret, ))
        return FileLeaf(orig_gfile.get_path())

    def get_description(self):
        return _("Move file back to original location")
    def get_icon_name(self):
        return "edit-undo"

class EmptyTrash (Action):
    rank_adjust = -1
    def __init__(self):
        Action.__init__(self, _("Empty Trash"))
    def activate(self, trash):
        gfile = Gio.File.new_for_uri(TRASH_URI)
        failed = []
        for info in gfile.enumerate_children("standard::*,trash::*",
                Gio.FileQueryInfoFlags.NONE, None):
            name = info.get_name()
            if not gfile.get_child(name).delete():
                failed.append(name)
        if failed:
            err = _("Could not delete files:\n    ")
            raise OperationError(err + '\n    '.join(failed))
    def get_icon_name(self):
        return "user-trash-full"

class TrashFile (Leaf):
    """A file in the trash. Represented object is a file info object"""
    def __init__(self, trash_uri, info):
        name = info.get_display_name()
        Leaf.__init__(self, info, name)
        self._trash_uri = trash_uri
    def get_actions(self):
        if self.get_orig_path():
            yield RestoreTrashedFile()
    def get_gfile(self):
        cur_gfile = Gio.File.new_for_uri(self._trash_uri).get_child(self.object.get_name())
        return cur_gfile
    def get_orig_path(self):
        try:
            orig_path = self.object.get_attribute_byte_string("trash::orig-path")
            return orig_path
        except AttributeError:
            pass
        return None

    def is_valid(self):
        return self.get_gfile().query_exists()

    def get_description(self):
        orig_path = self.get_orig_path()
        return utils.get_display_path_for_bytestring(orig_path) if orig_path \
                else None
    def get_gicon(self):
        return self.object.get_icon()
    def get_icon_name(self):
        return "text-x-generic"

class TrashContentSource (Source):
    def __init__(self, trash_uri, name):
        Source.__init__(self, name)
        self._trash_uri = trash_uri

    def is_dynamic(self):
        return True
    def get_items(self):
        gfile = Gio.File.new_for_uri(self._trash_uri)
        for info in gfile.enumerate_children("standard::*,trash::*",
                Gio.FileQueryInfoFlags.NONE, None):
            yield TrashFile(self._trash_uri, info)
    def should_sort_lexically(self):
        return True
    def get_gicon(self):
        return icons.get_gicon_for_file(self._trash_uri)

class SpecialLocation (Leaf):
    """ Base class for Special locations (in GIO/GVFS),
    such as trash:/// Here we assume they are all "directories"
    """
    def __init__(self, location, name=None, description=None, icon_name=None):
        """Special location with @location and
        @name. If unset, we find @name from filesystem
        @description is Leaf description"""
        gfile = Gio.File.new_for_uri(location)
        info = gfile.query_info(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME,
                Gio.FileQueryInfoFlags.NONE, None)
        name = (info.get_attribute_string(Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME) or location)
        Leaf.__init__(self, location, name)
        self.description = description
        self.icon_name = icon_name
    def get_actions(self):
        yield OpenTrash()
    def get_description(self):
        return self.description or self.object
    def get_gicon(self):
        # Get icon
        return icons.get_gicon_for_file(self.object)
    def get_icon_name(self):
        return "folder"

class Trash (SpecialLocation):
    def __init__(self, trash_uri, name=None):
        SpecialLocation.__init__(self, trash_uri, name=name)

    def has_content(self):
        return self.get_item_count()
    def content_source(self, alternate=False):
        return TrashContentSource(self.object, name=str(self))

    def get_actions(self):
        for action in SpecialLocation.get_actions(self):
            yield action
        if self.get_item_count():
            yield EmptyTrash()

    def get_item_count(self):
        gfile = Gio.File.new_for_uri(self.object)
        info = gfile.query_info(Gio.FILE_ATTRIBUTE_TRASH_ITEM_COUNT,
                Gio.FileQueryInfoFlags.NONE, None)
        return info.get_attribute_uint32(Gio.FILE_ATTRIBUTE_TRASH_ITEM_COUNT)

    def get_description(self):
        item_count = self.get_item_count()
        if not item_count:
            return _("Trash is empty")
        # proper translation of plural
        return ngettext("Trash contains one file",
            "Trash contains %(num)s files", item_count) % {"num": item_count}

class InvisibleSourceLeaf (SourceLeaf):
    """Hack to hide this source"""
    def is_valid(self):
        return False

class TrashSource (Source):
    def __init__(self):
        Source.__init__(self, _("Trash"))
    def get_items(self):
        try:
            yield Trash(TRASH_URI)
        except GLib.Error as exc:
            self.output_exc()
    def get_leaf_repr(self):
        return InvisibleSourceLeaf(self)
    def provides(self):
        yield SpecialLocation
    def get_icon_name(self):
        return "user-trash"

class OpenTrash(Open):
    def activate(self, obj, ctx=None):
        utils.show_url(obj.object)




import hashlib
import pickle
import os

from gi.repository import Gio, GLib

from kupfer import pretty
from kupfer import puid
from kupfer import conspickle

KUPFER_COMMAND_SHEBANG=b"#!/usr/bin/env kupfer-exec\n"

class ExecutionError (Exception):
    pass

def parse_kfcom_file(filepath):
    """Extract the serialized command inside @filepath

    The file must be executable (comparable to a shell script)
    >>> parse_kfcom_file(__file__)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ExecutionError: ... (not executable)

    Return commands triple
    """
    fobj = open(filepath, "rb")
    if not os.access(filepath, os.X_OK):
        raise ExecutionError(_('No permission to run "%s" (not executable)') %
                GLib.filename_display_basename(filepath))

    # strip shebang away
    data = fobj.read()
    if data.startswith(b"#!") and b"\n" in data:
        shebang, data = data.split(b"\n", 1)

    try:
        id_ = conspickle.BasicUnpickler.loads(data)
        command_object = puid.resolve_unique_id(id_)
    except pickle.UnpicklingError as err:
        raise ExecutionError("Could not parse: %s" % str(err))
    except Exception:
        raise ExecutionError('"%s" is not a saved command' %
                os.path.basename(filepath))
    if command_object is None:
        raise ExecutionError(_('Command in "%s" is not available') %
                GLib.filename_display_basename(filepath))

    try:
        return tuple(command_object.object)
    except (AttributeError, TypeError):
        raise ExecutionError('"%s" is not a saved command' %
                os.path.basename(filepath))
    finally:
        GLib.idle_add(update_icon, command_object, filepath)

def save_to_file(command_leaf, filename):
    fd = os.open(filename, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o777)
    wfile = os.fdopen(fd, "wb")
    try:
        wfile.write(KUPFER_COMMAND_SHEBANG)
        pickle.dump(puid.get_unique_id(command_leaf), wfile, protocol=3)
    finally:
        wfile.close()

def _write_thumbnail(gfile, pixbuf):
    uri = gfile.get_uri()
    hashname = hashlib.md5(uri.encode("utf-8")).hexdigest()
    thumb_dir = os.path.expanduser("~/.thumbnails/normal")
    if not os.path.exists(thumb_dir):
        os.makedirs(thumb_dir, 0o700)
    thumb_filename = os.path.join(thumb_dir, hashname + ".png")
    pixbuf.savev(thumb_filename, "png", [], [])
    return thumb_filename

def update_icon(kobj, filepath):
    "Give @filepath a custom icon taken from @kobj"
    icon_key = "metadata::custom-icon"

    gfile = Gio.File.new_for_path(filepath)
    finfo = gfile.query_info(icon_key, Gio.FileQueryInfoFlags.NONE, None)
    custom_icon_uri = finfo.get_attribute_string(icon_key)
    if custom_icon_uri and Gio.File.new_for_uri(custom_icon_uri).query_exists():
        return
    namespace = gfile.query_writable_namespaces() # FileAttributeInfoList
    if namespace.n_infos > 0:
        pretty.print_debug(__name__, "Updating icon for", filepath)
        thumb_filename = _write_thumbnail(gfile, kobj.get_pixbuf(128))
        try:
            gfile.set_attribute_string(icon_key,
                    Gio.File.new_for_path(thumb_filename).get_uri(),
                    Gio.FileQueryInfoFlags.NONE,
                    None)
        except GLib.GError:
            pretty.print_exc(__name__)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



__kupfer_name__ = _("Volumes and Disks")
__kupfer_sources__ = ("VolumesSource", )
__description__ = _("Mounted volumes and disks")
__version__ = "2017.2"
__author__ = ""

from gi.repository import Gio, GLib

from kupfer.objects import Action, Source, FileLeaf
from kupfer.obj.fileactions import Open, OpenTerminal
from kupfer import utils, uiutils


class Volume (FileLeaf):
    """
    The Volume class actually represents one instance
    of GIO's GMount (as long as it is mounted)
    """
    # NOTE: marking as non-serializable
    serializable = None

    def __init__(self, volume):
        self.volume = volume
        fil = self.volume.get_root()
        path = fil.get_path()
        super(Volume, self).__init__(obj=path, name=volume.get_name())
        self.kupfer_add_alias(fil.get_basename())

    def get_actions(self):
        yield Open()
        yield OpenTerminal()
        if self.volume.can_eject():
            yield Eject()
        elif self.volume.can_unmount():
            yield Unmount()

    def is_valid(self):
        vm = Gio.VolumeMonitor.get()
        return any(self.volume == v for v in vm.get_mounts())

    def get_description(self):
        return _("Volume mounted at %s") % \
                utils.get_display_path_for_bytestring(self.object)
    def get_gicon(self):
        return self.volume.get_icon()

    @classmethod
    def get_icon_name(self):
        return "drive-removable-media"

class Unmount (Action):
    def __init__(self, name=None):
        super(Unmount, self).__init__(name or _("Unmount"))

    def eject_callback(self, mount, async_result, ctx):
        try:
            mount.eject_with_operation_finish(async_result)
        except GLib.Error:
            ctx.register_late_error()
        else:
            self.success(mount.get_name())

    def unmount_callback(self, mount, async_result, ctx):
        try:
            mount.unmount_with_operation_finish(async_result)
        except GLib.Error:
            ctx.register_late_error()
        else:
            self.success(mount.get_name())

    def success(self, name):
        uiutils.show_notification(_("Unmount finished"),
                                  _('"%s" was successfully unmounted') % name,
                                  icon_name=Volume.get_icon_name())

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        if not leaf.is_valid():
            return
        vol = leaf.volume
        if vol.can_eject():
            vol.eject_with_operation(
                Gio.MountUnmountFlags.NONE,
                None,
                None,
                self.eject_callback,
                ctx)
        elif vol.can_unmount():
            vol.unmount_with_operation(
                Gio.MountUnmountFlags.NONE,
                None,
                None,
                self.unmount_callback, ctx)

    def get_description(self):
        return _("Unmount this volume")

    def get_icon_name(self):
        return "media-eject"

class Eject (Unmount):
    def __init__(self):
        super(Eject, self).__init__(_("Eject"))

    def get_description(self):
        return _("Unmount and eject this media")

class VolumesSource (Source):
    source_use_cache = False
    def __init__(self, name=_("Volumes and Disks")):
        super().__init__(name)

    def initialize(self):
        self.vm = Gio.VolumeMonitor.get()
        self.vm.connect("mount-added", self._update)
        self.vm.connect("mount-changed", self._update)
        self.vm.connect("mount-removed", self._update)

    def _update(self, *args):
        self.mark_for_update()
        GLib.timeout_add_seconds(1, lambda: self.mark_for_update())

    def finalize(self):
        del self.vm

    def get_items(self):
        # get_mounts gets all mounted removable media
        return (Volume(v) for v in self.vm.get_mounts())

    def get_description(self):
        return _("Mounted volumes and disks")
    def get_icon_name(self):
        return "drive-removable-media"
    def provides(self):
        yield Volume



__kupfer_name__ = _("Firefox Keywords")
__kupfer_sources__ = ("KeywordsSource", )
__kupfer_text_sources__ = ("KeywordSearchSource", )
__kupfer_actions__ = ("SearchWithEngine", )
__description__ = _("Search the web with Firefox keywords")
__version__ = "2020.1"
__author__ = ""

from contextlib import closing
import os
import sqlite3
import time
from urllib.parse import quote, urlparse

from kupfer import plugin_support
from kupfer.objects import Source, Action, Leaf
from kupfer.objects import TextLeaf, TextSource
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.objects import OpenUrl, RunnableLeaf
from kupfer import utils

from ._firefox_support import get_firefox_home_file

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "default",
        "label": _("Default for ?"),
        "type": str,
        "value": 'https://www.google.com/search?ie=UTF-8&q=%s',
    },
    {
        "key": "profile",
        "label": _("Firefox profile name or path"),
        "type": str,
        "value": "",
    },
)


def _url_domain(text):
    components = list(urlparse(text))
    domain = "".join(components[1:2])
    return domain

class Keyword(Leaf):
    def __init__(self, title, kw, url):
        title = title if title else _url_domain(url)
        name = "%s (%s)" % (kw, title)
        super().__init__(url, name)
        self.keyword = kw

    def _is_search(self):
        return "%s" in self.object

    def get_actions(self):
        if self._is_search():
            yield SearchFor()
        else:
            yield OpenUrl()

    def get_description(self):
        return self.object

    def get_icon_name(self):
        return "text-html"

    def get_text_representation(self):
        return self.object

class KeywordsSource(Source, FilesystemWatchMixin):
    instance = None
    def __init__(self):
        super().__init__(_("Firefox Keywords"))

    def initialize(self):
        KeywordsSource.instance = self
        profile = __kupfer_settings__["profile"]
        ff_home = get_firefox_home_file('', profile)
        self.monitor_token = self.monitor_directories(ff_home)

    def finalize(self):
        KeywordsSource.instance = None

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == 'lock'

    def _get_ffx3_bookmarks(self):
        """Query the firefox places bookmark database"""
        profile = __kupfer_settings__["profile"]
        fpath = get_firefox_home_file("places.sqlite", profile)
        if not (fpath and os.path.isfile(fpath)):
            return []

        fpath = fpath.replace("?", "%3f").replace("#", "%23")
        fpath = "file:" + fpath + "?immutable=1&mode=ro"

        for _ in range(2):
            try:
                self.output_debug("Reading bookmarks from", fpath)
                with closing(sqlite3.connect(fpath, uri=True,
                                             timeout=1)) as conn:
                    c = conn.cursor()
                    c.execute("""SELECT moz_places.url, moz_places.title,
                                  moz_keywords.keyword
                              FROM moz_places, moz_keywords
                              WHERE moz_places.id = moz_keywords.place_id
                              """)
                    return [Keyword(title, kw,  url) for url, title, kw in c]
            except sqlite3.Error as err:
                self.output_debug("Read bookmarks error:", str(err))
                # Something is wrong with the database
                # wait short time and try again
                time.sleep(1)
        self.output_exc()
        return []

    def get_items(self):
        seen_keywords = set()
        for kw in self._get_ffx3_bookmarks():
            seen_keywords.add(kw.keyword)
            yield kw

    def get_description(self):
        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield Keyword

class SearchWithEngine (Action):
    """TextLeaf -> SearchWithEngine -> Keyword"""
    action_accelerator = "s"
    def __init__(self):
        Action.__init__(self, _("Search With..."))

    def activate(self, leaf, iobj):
        url = iobj.object
        _do_search_engine(leaf.object, url)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield Keyword

    def valid_object(self, obj, for_item):
        return obj._is_search()

    def object_source(self, for_item=None):
        return KeywordsSource()

    def get_description(self):
        return _("Search the web with Firefox keywords")

    def get_icon_name(self):
        return "edit-find"

class SearchFor (Action):
    """Keyword -> SearchFor -> TextLeaf

    This is the opposite action to SearchWithEngine
    """
    action_accelerator = "s"
    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj):
        url = leaf.object
        terms = iobj.object
        _do_search_engine(terms, url)

    def item_types(self):
        yield Keyword

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def object_source(self, for_item):
        return TextSource(placeholder=_("Search Terms"))

    def valid_object(self, obj, for_item):
        # NOTE: Using exact class to skip subclasses
        return type(obj) == TextLeaf

    def get_description(self):
        return _("Search the web with Firefox keywords")

    def get_icon_name(self):
        return "edit-find"

class KeywordSearchSource(TextSource):
    def __init__(self):
        super().__init__(_("Firefox Keywords (?-source)"))

    def get_text_items(self, text):
        if not text.startswith("?"):
            return
        parts = text[1:].split(maxsplit=1)
        if len(parts) < 1:
            return
        query = parts[1] if len(parts) > 1 else ""
        for kw in KeywordsSource.instance.get_leaves():
            if kw._is_search() and kw.keyword == parts[0]:
                yield SearchWithKeyword(kw, query)
                return
        default = __kupfer_settings__['default'].strip()
        if default:
            if '%s' not in default:
                default += '%s'
            yield SearchWithKeyword(Keyword(None, "", default), text[1:])

    def get_description(self):
        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield SearchWithKeyword

    def get_rank(self):
        return 80

class SearchWithKeyword(RunnableLeaf):
    def __init__(self, keyword, text):
        super().__init__((keyword, text), _('Search for "%s"') % (text, ))

    def run(self):
        kw = self.keyword_leaf
        _do_search_engine(self.query, kw.object)

    @property
    def keyword_leaf(self):
        return self.object[0]

    @property
    def query(self):
        return self.object[1]

    def get_icon_name(self):
        return "web-browser"

    def get_description(self):
        return _("Search using %s") % self.keyword_leaf

    def get_text_representation(self):
        kw = self.keyword_leaf
        return _query_url(self.query, kw.object)

def _do_search_engine(terms, search_url, encoding="UTF-8"):
    """Show an url searching for @search_url with @terms"""
    utils.show_url(_query_url(terms, search_url))

def _query_url(terms, search_url):
    """Show an url searching for @search_url with @terms"""
    query_url = search_url.replace("%s", quote(terms))
    return query_url



__kupfer_name__ = _("Image Tools")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
        "Scale",
        "Rotate90",
        "Rotate270",
        "Autorotate",
    )
__description__ = _("Image transformation tools using 'convert' from ImageMagick")
__version__ = "2017.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

# since "path" is a very generic name, you often forget..
from os import path as os_path
import subprocess

from kupfer.objects import Action, FileLeaf, TextLeaf, OperationError
from kupfer import utils, pretty
from kupfer import runtimehelper

from kupfer.utils import SpawnError

def spawn_operation_err(argv):
    try:
        utils.spawn_async_raise(argv)
    except SpawnError as exc:
        raise OperationError(exc.args[0].message)

class Scale (Action):
    def __init__(self):
        Action.__init__(self, _("Scale..."))

    def has_result(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, obj, ctx):
        size = self._make_size(obj.object)
        fpath = leaf.object
        dirname = os_path.dirname(fpath)
        head, ext = os_path.splitext(os_path.basename(fpath))
        filename = "%s_%s%s" % (head, size, ext)
        dpath = utils.get_destpath_in_directory(dirname, filename)
        argv = ["convert", "-scale", ('%s' % size),  fpath, dpath]
        runtimehelper.register_async_file_result(ctx, dpath)
        spawn_operation_err(argv)
        return FileLeaf(dpath)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, item):
        # FIXME: Make this detection smarter
        root, ext = os_path.splitext(item.object)
        return ext.lower() in (".jpeg", ".jpg", ".png", ".gif")

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    @classmethod
    def _make_size(self, text):
        size = None
        # Allow leading =
        text = text.strip("= ")
        try:
            size = "%g" % float(text.strip())
        except ValueError:
            try:
                twoparts = text.split("x", 1)
                size = "%gx%g" % (float(twoparts[0].strip()),
                        float(twoparts[1].strip()))
            except ValueError:
                pass
        return size

    def valid_object(self, obj, for_item=None):
        return self._make_size(obj.object)

    def get_description(self):
        return _("Scale image to fit inside given pixel measure(s)")

class RotateBase (Action):
    def __init__(self, name, rotation):
        Action.__init__(self, name)
        self.rotation = rotation

    def has_result(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        fpath = leaf.object
        dirname = os_path.dirname(fpath)
        head, ext = os_path.splitext(os_path.basename(fpath))
        filename = "%s_%s%s" % (head, self.rotation, ext)
        dpath = utils.get_destpath_in_directory(dirname, filename)
        argv = ["jpegtran", "-copy", "all", "-rotate", self.rotation, "-outfile",
                dpath, fpath]
        runtimehelper.register_async_file_result(ctx, dpath)
        spawn_operation_err(argv)
        return FileLeaf(dpath)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, item):
        # FIXME: Make this detection smarter
        root, ext = os_path.splitext(item.object)
        return ext.lower() in (".jpeg", ".jpg")

class Rotate90 (RotateBase):
    def __init__(self):
        RotateBase.__init__(self, _("Rotate Clockwise"), "90")

    def get_icon_name(self):
        return "object-rotate-right"

class Rotate270 (RotateBase):
    def __init__(self):
        RotateBase.__init__(self, _("Rotate Counter-Clockwise"), "270")

    def get_icon_name(self):
        return "object-rotate-left"

class Autorotate (Action):
    def __init__(self):
        Action.__init__(self, _("Autorotate"))

    def has_result(self):
        return True

    def activate(self, leaf, obj=None):
        fpath = leaf.object
        argv = ['jhead', '-autorot', fpath]
        utils.spawn_async(argv)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, item):
        root, ext = os_path.splitext(item.object)
        if not ext.lower() in (".jpeg", ".jpg"):
            return False
        # Launch jhead to see if 1) it is installed, 2) Orientation nondefault
        try:
            cmdargs = ("jhead", item.object)
            proc = subprocess.Popen(cmdargs, stdout=subprocess.PIPE)
        except OSError:
            pretty.print_debug(__name__ , "Action %s needs 'jhead'" % self)
        else:
            out, err = proc.communicate()
            pretty.print_debug(__name__, "Running", cmdargs)
            return any(li.startswith("Orientation") for li in out.splitlines())

    def get_description(self):
        return _("Rotate JPEG (in-place) according to its EXIF metadata")




""" This plugin opens WhatsApp Web in the browser, and can send messages to numbers that are not scheduled 
"""

__kupfer_name__ = _("WhatsApp Web")
__kupfer_sources__ = ()
__kupfer_actions__ = ("WhatsApp",)
__description__ = _("""Open a new number in WhatsApp Web

The number format must respect:
"Country Code" + "Area Code" + "Number"

Example for a New York number: 12129999999
Country Code: 1
Area Code: 212
Number: 9999999

WhatsApp Web will open in the browser.

For help visit https://faq.whatsapp.com/general/chats/how-to-use-click-to-chat/

""")
__version__ = "1.0"
__author__ = "Leonardo Masuero <leom255255@gmail.com>"

from kupfer.objects import Action, TextLeaf
from kupfer import utils

class WhatsApp (Action):
	def __init__(self):
		Action.__init__(self, _("WhatsApp Web"))

	def activate(self, leaf):
		url_w = "https://web.whatsapp.com/send?phone="
		url_number = url_w + leaf.object
		utils.show_url(url_number)

	def item_types(self):
		yield TextLeaf

	def valid_for_item(self, leaf):
		try:
			text = leaf.object
			return text and int(text)
		except:
			pass

	def get_description(self):
		return _("Send a WhatsApp to a new number.")

	def get_icon_name(self):
		return "message-new"



"""
This module has a singleton Service for dbus callbacks,
and ensures there is only one unique service in the Session
"""

from gi.repository import GObject

try:
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)
    import dbus.service
    from dbus.gi_service import ExportedGObject

    dbus.mainloop.glib.threads_init()

# if dbus unavailable print the exception here
# but further actions (register) will fail without warning
    session_bus = dbus.Bus()
except (ImportError, dbus.exceptions.DBusException) as exc:
    session_bus = None
    print(exc)

from kupfer.ui import uievents

class AlreadyRunningError (Exception):
    """Service already available on the bus Exception"""
    pass

class NoConnectionError (Exception):
    """Not possible to establish connection
    for callbacks"""
    pass

server_name = "se.kaizer.kupfer"
interface_name = "se.kaizer.kupfer.Listener"
object_name = "/interface"

server_name_new = "io.github.kupferlauncher"
interface_name_new = "io.github.kupferlauncher.Listener"
object_name_new = "/io/github/kupferlauncher"

class Service (ExportedGObject):
    def __init__(self):
        """Create a new Kupfer service on the Session Bus

        Raises NoConnectionError, AlreadyRunningError
        """
        if not session_bus:
            raise NoConnectionError
        if session_bus.name_has_owner(server_name):
            raise AlreadyRunningError
        bus_name = dbus.service.BusName(server_name, bus=session_bus)
        super(Service, self).__init__(conn=session_bus, object_path=object_name,
                bus_name=bus_name)

    def unregister(self):
        if session_bus:
            session_bus.release_name(server_name)

    @dbus.service.method(interface_name)
    def Present(self):
        self.PresentOnDisplay("", "")

    @dbus.service.method(interface_name, in_signature="ay",
                         byte_arrays=True)
    def PresentWithStartup(self, notify_id):
        self.PresentOnDisplay("", notify_id)

    @dbus.service.method(interface_name, in_signature="ayay",
                         byte_arrays=True)
    def PresentOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("present", display, time)

    @dbus.service.method(interface_name)
    def ShowHide(self):
        self.emit("show-hide", "", 0)

    @dbus.service.method(interface_name, in_signature="s")
    def PutText(self, text):
        self.PutTextOnDisplay(text, "", "")

    @dbus.service.method(interface_name, in_signature="sayay",
                         byte_arrays=True)
    def PutTextOnDisplay(self, text, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-text", text, display, time)

    @dbus.service.method(interface_name, in_signature="as")
    def PutFiles(self, fileuris):
        self.PutFilesOnDisplay(fileuris, "", "")

    @dbus.service.method(interface_name, in_signature="asayay",
                         byte_arrays=True)
    def PutFilesOnDisplay(self, fileuris, display, notify_id):
        # files sent with dbus-send from kupfer have a custom comma
        # escape that we have to unescape here
        fileuris[:] = [f.replace("%%kupfercomma%%", ",") for f in fileuris]
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-files", fileuris, display, time)

    @dbus.service.method(interface_name, in_signature="s")
    def ExecuteFile(self, filepath):
        self.ExecuteFileOnDisplay(filepath, "", "")

    @dbus.service.method(interface_name, in_signature="sayay",
                         byte_arrays=True)
    def ExecuteFileOnDisplay(self, filepath, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("execute-file", filepath, display, time)

    @dbus.service.method(interface_name, in_signature="sayay",
                         byte_arrays=True)
    def RelayKeysFromDisplay(self, keystring, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("relay-keys", keystring, display, time)

    @dbus.service.method(interface_name, in_signature=None,
                         out_signature="as",
                         byte_arrays=True)
    def GetBoundKeys(self):
        from kupfer.ui import keybindings
        return keybindings.get_all_bound_keys()

    @dbus.service.signal(interface_name, signature="sb")
    def BoundKeyChanged(self, keystr, is_bound):
        pass

    @dbus.service.method(interface_name)
    def Quit(self):
        self.emit("quit")

# Signature: displayname, timestamp
GObject.signal_new("present", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: displayname, timestamp
GObject.signal_new("show-hide", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: text, displayname, timestamp
GObject.signal_new("put-text", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: filearray, displayname, timestamp
GObject.signal_new("put-files", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: fileuri, displayname, timestamp
GObject.signal_new("execute-file", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: ()
GObject.signal_new("quit", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())

# Signature: keystring, displayname, timestamp
GObject.signal_new("relay-keys", Service, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))


class ServiceNew (ExportedGObject):
    def __init__(self):
        """Create a new Kupfer service on the Session Bus

        Raises NoConnectionError, AlreadyRunningError
        """
        if not session_bus:
            raise NoConnectionError
        if session_bus.name_has_owner(server_name_new):
            raise AlreadyRunningError
        bus_name = dbus.service.BusName(server_name_new, bus=session_bus)
        super().__init__(conn=session_bus, object_path=object_name_new,
                bus_name=bus_name)

    def unregister(self):
        if session_bus:
            session_bus.release_name(server_name)

    @dbus.service.method(interface_name_new)
    def Present(self):
        self.PresentOnDisplay("", "")

    @dbus.service.method(interface_name_new, in_signature="ss")
    def PresentOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("present", display, time)

    @dbus.service.method(interface_name_new)
    def ShowHide(self):
        self.ShowHideOnDisplay("", "")

    @dbus.service.method(interface_name_new, in_signature="ss")
    def ShowHideOnDisplay(self, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("show-hide", display, time)

    @dbus.service.method(interface_name_new, in_signature="s")
    def PutText(self, text):
        self.PutTextOnDisplay(text, "", "")

    @dbus.service.method(interface_name_new, in_signature="sss")
    def PutTextOnDisplay(self, text, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-text", text, display, time)

    @dbus.service.method(interface_name_new, in_signature="as")
    def PutFiles(self, fileuris):
        self.PutFilesOnDisplay(fileuris, "", "")

    @dbus.service.method(interface_name_new, in_signature="asss")
    def PutFilesOnDisplay(self, fileuris, display, notify_id):
        # files sent with dbus-send from kupfer have a custom comma
        # escape that we have to unescape here
        fileuris[:] = [f.replace("%%kupfercomma%%", ",") for f in fileuris]
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("put-files", fileuris, display, time)

    @dbus.service.method(interface_name_new, in_signature="s")
    def ExecuteFile(self, filepath):
        self.ExecuteFileOnDisplay(filepath, "", "")

    @dbus.service.method(interface_name_new, in_signature="sss")
    def ExecuteFileOnDisplay(self, filepath, display, notify_id):
        with uievents.using_startup_notify_id(notify_id) as time:
            self.emit("execute-file", filepath, display, time)

    @dbus.service.method(interface_name_new)
    def Quit(self):
        self.emit("quit")

# Signature: displayname, timestamp
GObject.signal_new("present", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: displayname, timestamp
GObject.signal_new("show-hide", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: text, displayname, timestamp
GObject.signal_new("put-text", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: filearray, displayname, timestamp
GObject.signal_new("put-files", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: fileuri, displayname, timestamp
GObject.signal_new("execute-file", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))

# Signature: ()
GObject.signal_new("quit", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())


# Note ServiceNew has the same signals, but doesn't actually implement them
# all (yet). Needs new design for wayland keyrelay.
# Signature: keystring, displayname, timestamp
GObject.signal_new("relay-keys", ServiceNew, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_UINT))



# encoding: utf-8

def _unicode_truncate(ustr, length, encoding="UTF-8"):
    "Truncate @ustr to specific encoded byte length"
    bstr = ustr.encode(encoding)[:length]
    return bstr.decode(encoding, 'ignore')

def extract_title_body(text, maxtitlelen=60):
    """Prepare @text: Return a (title, body) tuple

    @text: A user-submitted paragraph or otherwise snippet of text. We
    try to detect an obvious title and then return the title and the
    following body. Otherwise we extract a title from the first words,
    and return the full text as body.

    @maxtitlelen: A unitless measure of approximate length of title.
    The default value yields a resulting title of approximately 60 ascii
    characters, or 20 asian characters.

    >>> extract_title_body("Short Text")
    ('Short Text', '')

    >>> title, body = extract_title_body(u"執筆方針については、項目名の付け方、"
    ...     "フォーマットや表記上の諸問題に関して多くの方針が存在している。")
    >>> print(title)
    執筆方針については、項目名の付け方、フォ
    >>> print(body)         # doctest: +ELLIPSIS
    執筆方針については、項目名の付け方、フォ...して多くの方針が存在している。
    """
    # if you don't make real tests, it's not not worth doing it at all.

    if not text.strip():
        return text, ""

    def split_first_line(text):
        """Take first non-empty line of text"""
        lines = iter(text.splitlines())
        for l in lines:
            l = l.strip()
            if not l:
                continue
            rest = "\n".join(lines)
            return l, rest

    # We use the UTF-8 encoding and truncate due to it:
    # this is a good heuristic for ascii vs "wide characters"
    # it results in taking fewer characters if they are asian, which
    # is exactly what we want
    def split_first_words(text, maxlen):
        text = text.lstrip()
        first_text = _unicode_truncate(text, maxlen)
        words = first_text.split()
        if len(words) > 3:
            words = words[:-1]
            first_words = " ".join(words[:-1])
            if text.startswith(first_words):
                first_text = first_words
        rest_text = text[len(first_text):]
        return first_text, rest_text

    firstline, rest = split_first_line(text)
    if len(firstline.encode("UTF-8")) > maxtitlelen:
        firstline, rest = split_first_words(text, maxtitlelen)
    else:
        return firstline, rest
    if rest.strip():
        return firstline, text
    else:
        return text, ""

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from kupfer.core import settings

def is_known_terminal_executable(exearg):
    setctl = settings.GetSettingsController()
    for id_, term in setctl.get_all_alternatives('terminal').items():
        if exearg == term["argv"][0]:
            return True
    return False


def get_configured_terminal():
    """
    Return the configured Terminal object
    """
    setctl = settings.GetSettingsController()
    return setctl.get_preferred_alternative('terminal')






import itertools
import operator
import os
import sys

from gi.repository import GLib, GObject

from kupfer.obj import base, sources, compose
from kupfer import pretty, scheduler
from kupfer import datatools
from kupfer.core import actioncompat
from kupfer.core import commandexec
from kupfer.core import execfile
from kupfer.core import pluginload
from kupfer.core import qfurl
from kupfer.core import search, learn
from kupfer.core import settings

from kupfer.core.sources import GetSourceController

# "Enums"
# Which pane
SourcePane, ActionPane, ObjectPane = (1,2,3)

# In two-pane or three-pane mode
SourceActionMode, SourceActionObjectMode = (1,2)

DATA_SAVE_INTERVAL_S = 3660

def identity(x):
    return x

def is_iterable(obj):
    return hasattr(obj, "__iter__")

def dress_leaves(seq, action):
    """yield items of @seq "dressed" by the source controller"""
    sc = GetSourceController()
    for itm in seq:
        sc.decorate_object(itm.object, action=action)
        yield itm

def peekfirst(seq):
    """This function will return (firstitem, iter)
    where firstitem is the first item of @seq or None if empty,
    and iter an equivalent copy of @seq
    """
    seq = iter(seq)
    for itm in seq:
        old_iter = itertools.chain((itm, ), seq)
        return (itm, old_iter)
    return (None, seq)

class Searcher (object):
    """
    This class searches KupferObjects efficiently, and
    stores searches in a cache for a very limited time (*)

    (*) As of this writing, the cache is used when the old key
    is a prefix of the search key.
    """

    def __init__(self):
        self._source_cache = {}
        self._old_key = None

    def search(self, sources, key, score=True, item_check=None, decorator=None):
        """
        @sources is a sequence listing the inputs, which should be
        Sources, TextSources or sequences of KupferObjects

        If @score, sort by rank.
        filters (with identity() as default):
            @item_check: Check items before adding to search pool
            @decorator: Decorate items before access

        Return (first, match_iter), where first is the first match,
        and match_iter an iterator to all matches, including the first match.
        """
        if not self._old_key or not key.startswith(self._old_key):
            self._source_cache.clear()
        self._old_key = key

        # General strategy: Extract a `list` from each source,
        # and perform ranking as in place operations on lists

        if not item_check: item_check = identity
        if not decorator: decorator = identity

        start_time = pretty.timing_start()
        match_lists = []
        for src in sources:
            fixedrank = 0
            can_cache = True
            rankables = None
            if is_iterable(src):
                items = item_check(src)
                can_cache = False
            else:
                # Look in source cache for stored rankables
                try:
                    rankables = self._source_cache[src]
                except KeyError:
                    try:
                        items = item_check(src.get_text_items(key))
                        fixedrank = src.get_rank()
                        can_cache = False
                    except AttributeError:
                        items = item_check(src.get_leaves())

            if rankables is None:
                rankables = search.make_rankables(items)

            if score:
                if fixedrank:
                    search.add_rank_objects(rankables, fixedrank)
                elif key:
                    search.score_objects(rankables, key)
                    search.bonus_objects(rankables, key)
                if can_cache:
                    self._source_cache[src] = rankables
            matches = rankables

            match_lists.append(matches)
        
        if score:
            matches = search.find_best_sort(match_lists)
        else:
            matches = itertools.chain(*match_lists)

        def as_set_iter(seq):
            key = operator.attrgetter("object")
            return datatools.UniqueIterator(seq, key=key)

        def valid_check(seq):
            """yield items of @seq that are valid"""
            for itm in seq:
                obj = itm.object
                if (not hasattr(obj, "is_valid")) or obj.is_valid():
                    yield itm

        # Check if the items are valid as the search
        # results are accessed through the iterators
        unique_matches = as_set_iter(matches)
        match, match_iter = peekfirst(decorator(valid_check(unique_matches)))
        pretty.timing_step(__name__, start_time, "ranked")
        return match, match_iter

    def rank_actions(self, objects, key, leaf, item_check=None, decorator=None):
        """
        rank @objects, which should be a sequence of KupferObjects,
        for @key, with the action ranker algorithm.

        @leaf is the Leaf the action is going to be invoked on

        Filters and return value like .score().
        """
        if not item_check: item_check = identity
        if not decorator: decorator = identity

        rankables = search.make_rankables(item_check(objects))
        if key:
            search.score_objects(rankables, key)
            matches = search.bonus_actions(rankables, key)
        else:
            matches = search.score_actions(rankables, leaf)
        matches = sorted(matches, key=operator.attrgetter("rank"), reverse=True)

        match, match_iter = peekfirst(decorator(matches))
        return match, match_iter

class Pane (GObject.GObject):
    """
    signals:
        search-result (match, match_iter, context)
    """
    __gtype_name__ = "Pane"
    def __init__(self):
        super(Pane, self).__init__()
        self.selection = None
        self.latest_key = None
        self.outstanding_search = -1
        self.outstanding_search_id = -1
        self.searcher = Searcher()

    def select(self, item):
        self.selection = item
    def get_selection(self):
        return self.selection
    def reset(self):
        self.selection = None
    def get_latest_key(self):
        return self.latest_key
    def get_can_enter_text_mode(self):
        return False
    def get_should_enter_text_mode(self):
        return False
    def emit_search_result(self, match, match_iter, context):
        self.emit("search-result", match, match_iter, context)

GObject.signal_new("search-result", Pane, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, 
        GObject.TYPE_PYOBJECT))

class LeafPane (Pane, pretty.OutputMixin):
    __gtype_name__ = "LeafPane"

    def __init__(self):
        super(LeafPane, self).__init__()
        self.source_stack = []
        self.source = None
        self.object_stack = []

    def select(self, item):
        assert item is None or isinstance(item, base.Leaf), \
                    "New selection for object pane is not a Leaf!"
        super().select(item)

    def _load_source(self, src):
        """Try to get a source from the SourceController,
        if it is already loaded we get it from there, else
        returns @src"""
        sc = GetSourceController()
        return sc.get_canonical_source(src)

    def get_source(self):
        return self.source

    def source_rebase(self, src):
        self.source_stack = []
        self.source = self._load_source(src)
        self.refresh_data()

    def push_source(self, src):
        self.source_stack.append(self.source)
        self.source = self._load_source(src)
        self.refresh_data()

    def pop_source(self):
        """Return True if succeeded"""
        if not len(self.source_stack):
            return False
        self.source = self.source_stack.pop()
        return True

    def is_at_source_root(self):
        """Return True if we have no source stack"""
        return not self.source_stack

    def object_stack_push(self, obj):
        self.object_stack.append(obj)

    def object_stack_pop(self):
        return self.object_stack.pop()

    def get_can_enter_text_mode(self):
        return self.is_at_source_root()

    def get_should_enter_text_mode(self):
        return False

    def refresh_data(self):
        self.emit("new-source", self.source)

    def browse_up(self):
        """Try to browse up to previous sources, from current
        source"""
        succ = self.pop_source()
        if not succ:
            if self.source.has_parent():
                self.source_rebase(self.source.get_parent())
                succ = True
        if succ:
            self.refresh_data()
        return succ

    def browse_down(self, alternate=False):
        """Browse into @leaf if it's possible
        and save away the previous sources in the stack
        if @alternate, use the Source's alternate method"""
        leaf = self.get_selection()
        if leaf and leaf.has_content():
            self.push_source(leaf.content_source(alternate=alternate))
            return True
        return False

    def reset(self):
        """Pop all sources and go back to top level"""
        Pane.reset(self)
        while self.pop_source():
            pass
        self.refresh_data()

    def soft_reset(self):
        Pane.reset(self)
        while self.pop_source():
            pass
        return self.source

    def search(self, key="", context=None, text_mode=False):
        """
        filter for action @item
        """
        self.latest_key = key
        sources = [ self.get_source() ] if not text_mode else []
        if key and self.is_at_source_root():
            # Only use text sources when we are at root catalog
            sc = GetSourceController()
            textsrcs = sc.get_text_sources()
            sources.extend(textsrcs)

        decorator = lambda seq: dress_leaves(seq, action=None)
        match, match_iter = self.searcher.search(sources, key, score=bool(key),
                decorator=decorator)
        self.emit_search_result(match, match_iter, context)

GObject.signal_new("new-source", LeafPane, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_PYOBJECT,))

class PrimaryActionPane (Pane):
    def __init__(self):
        super(PrimaryActionPane, self).__init__()
        self.set_item(None)

    def select(self, item):
        assert not item or isinstance(item, base.Action), \
                "Selection in action pane is not an Action!"
        super().select(item)

    def set_item(self, item):
        """Set which @item we are currently listing actions for"""
        self.current_item = item
        self._action_valid_cache = {}

    def search(self, key="", context=None, text_mode=False):
        """Search: Register the search method in the event loop

        using @key, promising to return
        @context in the notification about the result, having selected
        @item in SourcePane

        If we already have a call to search, we remove the "source"
        so that we always use the most recently requested search."""

        self.latest_key = key
        leaf = self.current_item
        actions = actioncompat.actions_for_item(leaf, GetSourceController())

        def is_valid_cached(action):
            """Check if @action is valid for current item"""
            cache = self._action_valid_cache
            valid = cache.get(action)
            if valid is None:
                valid = actioncompat.action_valid_for_item(action, leaf)
                cache[action] = valid
            return valid

        def valid_decorator(seq):
            """Check if actions are valid before access"""
            for obj in seq:
                if is_valid_cached(obj.object):
                    yield obj

        match, match_iter = self.searcher.rank_actions(actions, key, leaf,
                decorator=valid_decorator)
        self.emit_search_result(match, match_iter, context)

class SecondaryObjectPane (LeafPane):
    __gtype_name__ = "SecondaryObjectPane"
    def __init__(self):
        LeafPane.__init__(self)
        self.current_item = None
        self.current_action = None

    def reset(self):
        LeafPane.reset(self)
        self.searcher = Searcher()

    def set_item_and_action(self, item, act):
        self.current_item = item
        self.current_action = act
        if item and act:
            ownsrc, use_catalog = actioncompat.iobject_source_for_action(act, item)
            if ownsrc and not use_catalog:
                self.source_rebase(ownsrc)
            else:
                extra_sources = [ownsrc] if ownsrc else ()
                sc = GetSourceController()
                self.source_rebase(sc.root_for_types(act.object_types(), extra_sources))
        else:
            self.reset()

    def get_can_enter_text_mode(self):
        """Check if there are any reasonable text sources for this action"""
        atroot = self.is_at_source_root()
        types = tuple(self.current_action.object_types())
        sc = GetSourceController()
        textsrcs = sc.get_text_sources()
        return (atroot and
            any(sc.good_source_for_types(s, types) for s in textsrcs))

    def get_should_enter_text_mode(self):
        return (self.is_at_source_root() and
                hasattr(self.get_source(), "get_text_items"))

    def search(self, key="", context=None, text_mode=False):
        """
        filter for action @item
        """
        self.latest_key = key
        sources = []
        if not text_mode or hasattr(self.get_source(), "get_text_items"):
            sources.append(self.get_source())
        if key and self.is_at_source_root():
            # Only use text sources when we are at root catalog
            sc = GetSourceController()
            textsrcs = sc.get_text_sources()
            sources.extend(textsrcs)

        item_check = actioncompat.iobjects_valid_for_action(self.current_action,
                self.current_item)
        decorator = lambda seq: dress_leaves(seq, action=self.current_action)

        match, match_iter = self.searcher.search(sources, key, score=True,
                item_check=item_check, decorator=decorator)
        self.emit_search_result(match, match_iter, context)

class DataController (GObject.GObject, pretty.OutputMixin):
    """
    Sources <-> Actions controller

    The data controller must be created before main program commences,
    so it can register itself at the scheduler correctly.
    """
    __gtype_name__ = "DataController"

    def __init__(self):
        super(DataController, self).__init__()

        self.source_pane = LeafPane()
        self.object_pane = SecondaryObjectPane()
        self.source_pane.connect("new-source", self._new_source)
        self.object_pane.connect("new-source", self._new_source)
        self.action_pane = PrimaryActionPane()
        self._panectl_table = {
            SourcePane : self.source_pane,
            ActionPane : self.action_pane,
            ObjectPane : self.object_pane,
            }
        for pane, ctl in self._panectl_table.items():
            ctl.connect("search-result", self._pane_search_result, pane)
        self.mode = None
        self._search_ids = itertools.count(1)
        self._latest_interaction = -1
        self._execution_context = commandexec.DefaultActionExecutionContext()
        self._execution_context.connect("command-result",
                self._command_execution_result)
        self._execution_context.connect("late-command-result",
                self._late_command_execution_result)

        self._save_data_timer = scheduler.Timer()

        sch = scheduler.GetScheduler()
        sch.connect("load", self._load)
        sch.connect("display", self._display)
        sch.connect("finish", self._finish)

    def register_text_sources(self, plugin_id, srcs):
        """Pass in text sources as @srcs

        we register text sources """
        sc = GetSourceController()
        sc.add_text_sources(plugin_id, srcs)
    
    def register_action_decorators(self, plugin_id, actions):
        # Keep a mapping: Decorated Leaf Type -> List of actions
        decorate_types = {}
        for action in actions:
            for appl_type in action.item_types():
                decorate_types.setdefault(appl_type, []).append(action)
        if not decorate_types:
            return
        sc = GetSourceController()
        sc.add_action_decorators(plugin_id, decorate_types)

    def register_content_decorators(self, plugin_id, contents):
        """
        Register the sequence of classes @contents as
        potential content decorators. Classes not conforming to
        the decoration protocol (most importantly, ``.decorates_type()``)
        will be skipped
        """
        # Keep a mapping:
        # Decorated Leaf Type -> Set of content decorator types
        decorate_item_types = {}
        for c in contents:
            try:
                applies = c.decorates_type()
            except AttributeError:
                continue
            decorate_item_types.setdefault(applies, set()).add(c)
        if not decorate_item_types:
            return
        sc = GetSourceController()
        sc.add_content_decorators(plugin_id, decorate_item_types)

    def register_action_generators(self, plugin_id, generators):
        sc = GetSourceController()
        for generator in generators:
            sc.add_action_generator(plugin_id, generator)

    def _load(self, sched):
        """Begin Data Controller work when we get application 'load' signal

        Load the data model from saved configuration and caches
        """
        setctl = settings.GetSettingsController()
        setctl.connect("plugin-enabled-changed", self._plugin_enabled)
        setctl.connect("plugin-toplevel-changed", self._plugin_catalog_changed)

        self._load_all_plugins()
        D_s, d_s = self._get_directory_sources()
        sc = GetSourceController()
        sc.add(None, D_s, toplevel=True)
        sc.add(None, d_s, toplevel=False)
        sc.initialize()
        learn.load()

    def _display(self, sched):
        self._reload_source_root()
        self._save_data_timer.set(DATA_SAVE_INTERVAL_S, self._save_data)

    def _get_directory_sources(self):
        """
        Return a tuple of S_sources, s_sources for
        directory sources directly included and for
        catalog inclusion respectively
        """

        s_sources = []
        S_sources = []

        setctl = settings.GetSettingsController()
        source_config = setctl.get_config

        def dir_source(opt):
            return sources.DirectorySource(opt)

        def file_source(opt, depth=1):
            abs = os.path.abspath(os.path.expanduser(opt))
            return sources.FileSource((abs,), depth)

        for coll, direct in zip((s_sources, S_sources), (False, True)):
            for item in setctl.get_directories(direct):
                if not os.path.isdir(item):
                    continue
                coll.append(dir_source(item))

        dir_depth = source_config("DeepDirectories", "Depth")

        for item in source_config("DeepDirectories","Catalog"):
            s_sources.append(file_source(item, dir_depth))
        for item in source_config("DeepDirectories", "Direct"):
            S_sources.append(file_source(item, dir_depth))

        return S_sources, s_sources

    def _load_all_plugins(self):
        """
        Insert all plugin sources into the catalog
        """
        from kupfer.core import plugins

        setctl = settings.GetSettingsController()
        for item in sorted(plugins.get_plugin_ids()):
            if not setctl.get_plugin_enabled(item):
                continue
            sources = self._load_plugin(item)
            self._insert_sources(item, sources, initialize=False)

    def _load_plugin(self, plugin_id):
        """
        Load @plugin_id, register all its Actions, Content and TextSources.
        Return its sources.
        """
        with pluginload.exception_guard(plugin_id):
            plugin = pluginload.load_plugin(plugin_id)
            self.register_text_sources(plugin_id, plugin.text_sources)
            self.register_action_decorators(plugin_id, plugin.action_decorators)
            self.register_content_decorators(plugin_id, plugin.content_decorators)
            self.register_action_generators(plugin_id, plugin.action_generators)
            return set(plugin.sources)
        return set()

    def _plugin_enabled(self, setctl, plugin_id, enabled):
        from kupfer.core import plugins
        if enabled and not plugins.is_plugin_loaded(plugin_id):
            sources = self._load_plugin(plugin_id)
            self._insert_sources(plugin_id, sources, initialize=True)
        elif not enabled:
            self._remove_plugin(plugin_id)

    def _remove_plugin(self, plugin_id):
        sc = GetSourceController()
        if sc.remove_objects_for_plugin_id(plugin_id):
            self._reload_source_root()
        pluginload.remove_plugin(plugin_id)

    def _reload_source_root(self):
        self.output_debug("Reloading source root")
        sc = GetSourceController()
        self.source_pane.source_rebase(sc.root)

    def _plugin_catalog_changed(self, setctl, plugin_id, toplevel):
        self._reload_source_root()

    def _insert_sources(self, plugin_id, sources, initialize=True):
        if not sources:
            return
        sc = GetSourceController()
        setctl = settings.GetSettingsController()
        for src in sources:
            is_toplevel = setctl.get_source_is_toplevel(plugin_id, src)
            sc.add(plugin_id, (src, ), toplevel=is_toplevel,
                   initialize=initialize)
        if initialize:
            self._reload_source_root()

    def _finish(self, sched):
        "Close down the data model, save user data, and write caches to disk"
        GetSourceController().finalize()
        self._save_data(final_invocation=True)
        self.output_info("Saving cache...")
        GetSourceController().save_cache()

    def _save_data(self, final_invocation=False):
        """Save Learning data and User's configuration data in sources
        (Recurring timer)
        """
        self.output_info("Saving data...")
        learn.save()
        GetSourceController().save_data()
        if not final_invocation:
            self._save_data_timer.set(DATA_SAVE_INTERVAL_S, self._save_data)

    def _new_source(self, ctr, src):
        if ctr is self.source_pane:
            pane = SourcePane
        elif ctr is self.object_pane:
            pane = ObjectPane
        root = ctr.is_at_source_root()
        self.emit("source-changed", pane, src, root)

    def reset(self):
        self.source_pane.reset()
        self.action_pane.reset()

    def soft_reset(self, pane):
        if pane is ActionPane:
            return
        panectl = self._panectl_table[pane]
        return panectl.soft_reset()

    def cancel_search(self, pane=None):
        """Cancel any outstanding search, or the search for @pane"""
        panes = (pane, ) if pane else iter(self._panectl_table)
        for pane in panes:
            ctl = self._panectl_table[pane]
            if ctl.outstanding_search > 0:
                GLib.source_remove(ctl.outstanding_search)
                ctl.outstanding_search = -1

    def search(self, pane, key="", context=None, interactive=False, lazy=False,
            text_mode=False):
        """Search: Register the search method in the event loop

        Will search in @pane's base using @key, promising to return
        @context in the notification about the result.

        if @interactive, the search result will return immediately
        if @lazy, will slow down search result reporting
        """

        self.cancel_search(pane)
        self._latest_interaction = self._execution_context.last_command_id
        ctl = self._panectl_table[pane]
        ctl.outstanding_search_id = next(self._search_ids)
        wrapcontext = (ctl.outstanding_search_id, context)
        if interactive:
            ctl.search(key, wrapcontext, text_mode)
        else:
            timeout = 300 if lazy else 0 if not key else 50//len(key)

            def ctl_search(*args):
                ctl.outstanding_search = -1
                return ctl.search(*args)
            ctl.outstanding_search = GLib.timeout_add(timeout, ctl_search,
                    key, wrapcontext, text_mode)

    def _pane_search_result(self, panectl, match,match_iter, wrapcontext, pane):
        search_id, context = wrapcontext
        if not search_id is panectl.outstanding_search_id:
            self.output_debug("Skipping late search", match, context)
            return True
        self.emit("search-result", pane, match, match_iter, context)

    def select(self, pane, item):
        """Select @item in @pane to self-update
        relevant places"""
        # If already selected, do nothing
        panectl = self._panectl_table[pane]
        if item == panectl.get_selection():
            return
        self.cancel_search()
        panectl.select(item)
        if pane is SourcePane:
            # populate actions
            citem = self._get_pane_object_composed(self.source_pane)
            self.action_pane.set_item(citem)
            self.search(ActionPane, interactive=True)
            if self.mode == SourceActionObjectMode:
                self.object_stack_clear(ObjectPane)
                self._populate_third_pane()
        elif pane is ActionPane:
            self.object_stack_clear(ObjectPane)
            if item and item.requires_object():
                newmode = SourceActionObjectMode
            else:
                newmode = SourceActionMode
            if newmode != self.mode:
                self.mode = newmode
                self.emit("mode-changed", self.mode, item)
            if self.mode == SourceActionObjectMode:
                self._populate_third_pane()
        elif pane is ObjectPane:
            pass

    def _populate_third_pane(self):
        citem = self._get_pane_object_composed(self.source_pane)
        action = self.action_pane.get_selection()
        self.object_pane.set_item_and_action(citem, action)
        self.search(ObjectPane, lazy=True)

    def get_can_enter_text_mode(self, pane):
        panectl = self._panectl_table[pane]
        return panectl.get_can_enter_text_mode()

    def get_should_enter_text_mode(self, pane):
        panectl = self._panectl_table[pane]
        return panectl.get_should_enter_text_mode()

    def validate(self):
        """Check if all selected items are still valid
        (for example after being spawned again, old item
        still focused)

        This will trigger .select() with None if items
        are not valid..
        """
        def valid_check(obj):
            return not (hasattr(obj, "is_valid") and not obj.is_valid())

        for pane, panectl in self._panectl_table.items():
            sel = panectl.get_selection()
            if not valid_check(sel):
                self.emit("pane-reset", pane, None)
                self.select(pane, None)
            if self._has_object_stack(pane):
                new_stack = [o for o in panectl.object_stack if valid_check(o)]
                if new_stack != panectl.object_stack:
                    self._set_object_stack(pane, new_stack)

    def browse_up(self, pane):
        """Try to browse up to previous sources, from current
        source"""
        if pane is SourcePane:
            return self.source_pane.browse_up()
        if pane is ObjectPane:
            return self.object_pane.browse_up()
    
    def browse_down(self, pane, alternate=False):
        """Browse into @leaf if it's possible
        and save away the previous sources in the stack
        if @alternate, use the Source's alternate method"""
        if pane is ActionPane:
            return
        # record used object if we browse down
        panectl = self._panectl_table[pane]
        sel, key = panectl.get_selection(), panectl.get_latest_key()
        if panectl.browse_down(alternate=alternate):
            learn.record_search_hit(sel, key)

    def activate(self, ui_ctx):
        """
        Activate current selection

        @ui_ctx: GUI environment context object
        """
        leaf, action, sobject = self._get_current_command_objects()

        # register search to learning database
        learn.record_search_hit(leaf, self.source_pane.get_latest_key())
        learn.record_search_hit(action, self.action_pane.get_latest_key())
        if sobject and self.mode is SourceActionObjectMode:
            learn.record_search_hit(sobject, self.object_pane.get_latest_key())

        try:
            ctx = self._execution_context
            res, ret = ctx.run(leaf, action, sobject, ui_ctx=ui_ctx)
        except commandexec.ActionExecutionError:
            self.output_exc()
            return

        if res not in commandexec.RESULTS_SYNC:
            self.emit("launched-action")

    def execute_file(self, filepath, ui_ctx, on_error):
        try:
            cmd_objs = execfile.parse_kfcom_file(filepath)
            ctx = self._execution_context
            ctx.run(*cmd_objs, ui_ctx=ui_ctx)
            return True
        except commandexec.ActionExecutionError:
            self.output_exc()
            return
        except execfile.ExecutionError:
            on_error(sys.exc_info())
            return False

    def _insert_object(self, pane, obj):
        "Insert @obj in @pane: prepare the object, then emit pane-reset"
        self._decorate_object(obj)
        self.emit("pane-reset", pane, search.wrap_rankable(obj))

    def _decorate_object(self, *objects):
        sc = GetSourceController()
        for obj in objects:
            sc.decorate_object(obj)

    def insert_objects(self, pane, objects):
        "Select @objects in @pane"
        if pane != SourcePane:
            raise ValueError("Can only insert in first pane")
        self._decorate_object(objects[:-1])
        self._set_object_stack(pane, objects[:-1])
        self._insert_object(pane, objects[-1])

    def _command_execution_result(self, ctx, result_type, ret, uictx):
        if result_type == commandexec.RESULT_SOURCE:
            self.object_stack_clear_all()
            self.source_pane.push_source(ret)
        elif result_type == commandexec.RESULT_OBJECT:
            self.object_stack_clear_all()
            self._insert_object(SourcePane, ret)
        else:
            return
        self.emit("command-result", result_type, uictx)

    def _late_command_execution_result(self, ctx, id_, result_type, ret, uictx):
        "Receive late command result"
        if self._latest_interaction < id_:
            self._command_execution_result(ctx, result_type, ret, uictx)

    def find_object(self, url):
        """Find object with URI @url and select it in the first pane"""
        sc = GetSourceController()
        qf = qfurl.qfurl(url=url)
        found = qf.resolve_in_catalog(sc.sources)
        if found and not found == self.source_pane.get_selection():
            self._insert_object(SourcePane, found)

    def mark_as_default(self, pane):
        """
        Make the object selected on @pane as default
        for the selection in previous pane.
        """
        if pane is SourcePane or pane is ObjectPane:
            raise RuntimeError("Setting default on pane 1 or 3 not supported")
        obj = self.source_pane.get_selection()
        act = self.action_pane.get_selection()
        assert obj and act
        learn.set_correlation(act, obj)

    def get_object_has_affinity(self, pane):
        """
        Return ``True`` if we have any recorded affinity
        for the object selected in @pane
        """
        panectl = self._panectl_table[pane]
        selection = panectl.get_selection()
        if not selection:
            return None
        return learn.get_object_has_affinity(selection)

    def erase_object_affinity(self, pane):
        """
        Erase all learned and configured affinity for
        the selection of @pane
        """
        panectl = self._panectl_table[pane]
        selection = panectl.get_selection()
        if not selection:
            return None
        return learn.erase_object_affinity(selection)

    def compose_selection(self):
        leaf, action, iobj = self._get_current_command_objects()
        if leaf is None:
            return
        self.object_stack_clear_all()
        obj = compose.ComposedLeaf(leaf, action, iobj)
        self._insert_object(SourcePane, obj)

    def _get_pane_object_composed(self, pane):
        objects = list(pane.object_stack)
        sel = pane.get_selection()
        if sel and sel not in objects:
            objects.append(sel)
        if not objects:
            return None
        elif len(objects) == 1:
            return objects[0]
        else:
            return compose.MultipleLeaf(objects)

    def _get_current_command_objects(self):
        """
        Return a tuple of current (obj, action, iobj)
        """
        objects = self._get_pane_object_composed(self.source_pane)
        action = self.action_pane.get_selection()
        if objects is None or action is None:
            return (None, None, None)
        iobjects = self._get_pane_object_composed(self.object_pane)
        if self.mode == SourceActionObjectMode:
            if not iobjects:
                return (None, None, None)
        else:
            iobjects = None
        return (objects, action, iobjects)

    def _has_object_stack(self, pane):
        return pane in (SourcePane, ObjectPane)

    def _set_object_stack(self, pane, newstack):
        panectl = self._panectl_table[pane]
        panectl.object_stack[:] = list(newstack)
        self.emit("object-stack-changed", pane)

    def object_stack_push(self, pane, object_):
        """
        Push @object_ onto the stack
        """
        if not self._has_object_stack(pane):
            return
        panectl = self._panectl_table[pane]
        if object_ not in panectl.object_stack:
            panectl.object_stack_push(object_)
            self.emit("object-stack-changed", pane)
        return True

    def object_stack_pop(self, pane):
        if not self._has_object_stack(pane):
            return
        panectl = self._panectl_table[pane]
        obj = panectl.object_stack_pop()
        self._insert_object(pane, obj)
        self.emit("object-stack-changed", pane)
        return True

    def object_stack_clear(self, pane):
        if not self._has_object_stack(pane):
            return
        panectl = self._panectl_table[pane]
        panectl.object_stack[:] = []
        self.emit("object-stack-changed", pane)

    def object_stack_clear_all(self):
        """
        Clear the object stack for all panes
        """
        for pane in self._panectl_table:
            self.object_stack_clear(pane)

    def get_object_stack(self, pane):
        if not self._has_object_stack(pane):
            return ()
        panectl = self._panectl_table[pane]
        return panectl.object_stack

# pane cleared or set with new item
# pane, item
GObject.signal_new("pane-reset", DataController, GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN, (GObject.TYPE_INT, GObject.TYPE_PYOBJECT,))

# pane, match, iter to matches, context
GObject.signal_new("search-result", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_INT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT))

GObject.signal_new("source-changed", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (int, object, bool))

# mode, None(?)
GObject.signal_new("mode-changed", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_INT, GObject.TYPE_PYOBJECT,))

# object stack update signal
# arguments: pane
GObject.signal_new("object-stack-changed", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_INT, ))
# when an command returned a result
# arguments: result type, gui_context
GObject.signal_new("command-result", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_INT, GObject.TYPE_PYOBJECT))

# when an action was launched
# arguments: none
GObject.signal_new("launched-action", DataController, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())






__kupfer_name__ = _("Applications")
__kupfer_sources__ = ("AppSource", )
__kupfer_actions__ = (
        "OpenWith",
        "SetDefaultApplication",
        "ResetAssociations",
    )
__description__ = _("All applications and preferences")
__version__ = "2017.3"
__author__ = ""

from gi.repository import Gio

from kupfer.objects import Action, Source, AppLeaf, FileLeaf
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer import config, plugin_support
from kupfer.weaklib import gobject_connect_weakly

_ALTERNATIVES = (
    "",
    "Cinnamon",
    "EDE",
    "GNOME",
    "KDE",
    "LXDE",
    "LXQt",
    "MATE",
    "Pantheon",
    "ROX",
    "Razor",
    "TDE",
    "Unity",
    "XFCE",
)

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "desktop_type",
        "label": _("Applications for Desktop Environment"),
        "type": str,
        "value": "",
        "alternatives": _ALTERNATIVES,
    },
    {
        "key" : "desktop_filter",
        "label": _("Use Desktop Filter"),
        "type": bool,
        "value": True,
    },
)

# Gio.AppInfo / Desktop Item nodisplay vs hidden:
# NoDisplay: Don't show this in program menus
# Hidden: Disable/never use at all

WHITELIST_IDS = frozenset([
    # we think that these are useful to show
    "eog.desktop",
    "evince.desktop",
    "gnome-about.desktop",
    "gstreamer-properties.desktop",
    "notification-properties.desktop",
    "shotwell-viewer.desktop",
    ])
BLACKLIST_IDS = frozenset([
    "nautilus-home.desktop",
])

class AppSource (Source, FilesystemWatchMixin):
    """
    Applications source

    This Source contains all user-visible applications (as given by
    the desktop files)
    """
    def __init__(self, name=None):
        super().__init__(name or _("Applications"))

    def initialize(self):
        application_dirs = config.get_data_dirs("", "applications")
        self.monitor_token = self.monitor_directories(*application_dirs)
        gobject_connect_weakly(__kupfer_settings__, "plugin-setting-changed",
                               self._on_setting_change)

    def _on_setting_change(self, *_args):
        self.mark_for_update()

    @classmethod
    def should_show(cls, app_info, desktop_type, use_filter):
        if app_info.get_nodisplay():
            return False
        if not use_filter:
            return True
        if desktop_type == "":
            return app_info.should_show()
        else:
            return app_info.get_show_in(desktop_type)

    def get_items(self):
        use_filter = __kupfer_settings__["desktop_filter"]
        desktop_type = __kupfer_settings__["desktop_type"]

        # Add this to the default

        for item in Gio.app_info_get_all():
            id_ = item.get_id()
            if id_ in WHITELIST_IDS or (
                self.should_show(item, desktop_type, use_filter)
                and not id_ in BLACKLIST_IDS):
                yield AppLeaf(item)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("All applications and preferences")

    def get_icon_name(self):
        return "applications-office"
    def provides(self):
        yield AppLeaf

class OpenWith(Action):
    action_accelerator = "w"
    def __init__(self):
        super().__init__(_("Open With..."))

    def _activate(self, app_leaf, paths, ctx):
        app_leaf.launch(paths=paths, ctx=ctx)

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        self.activate_multiple((leaf, ), (iobj, ), ctx)

    def activate_multiple(self, objects, iobjects, ctx):
        # for each application, launch all the files
        for iobj_app in iobjects:
            self._activate(iobj_app, [L.object for L in objects], ctx)
        return
        for iobj_app in iobjects:
            for L in objects:
                ct = L.get_content_type()
                if ct:
                    iobj_app.object.set_as_last_used_for_type(ct)

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, leaf):
        return AppsAll()

    def object_source_and_catalog(self, leaf):
        return True

    def valid_object(self, iobj, for_item):
        return iobj.object.supports_files() or iobj.object.supports_uris()

    def get_description(self):
        return _("Open with any application")

class SetDefaultApplication (Action):
    def __init__(self):
        super().__init__(_("Set Default Application..."))

    def activate(self, leaf, obj):
        desktop_item = obj.object
        desktop_item.set_as_default_for_type(leaf.get_content_type())

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, leaf):
        return AppsAll()

    def object_source_and_catalog(self, leaf):
        return True

    def valid_object(self, iobj, for_item):
        return iobj.object.supports_files() or iobj.object.supports_uris()

    def get_description(self):
        return _("Set default application to open this file type")

class AppsAll(Source):
    def __init__(self):
        super().__init__(_("Applications"))

    def get_items(self):
        use_filter = __kupfer_settings__["desktop_filter"]
        desktop_type = __kupfer_settings__["desktop_type"]

        # Get all apps; this includes those only configured for
        # opening files with.
        for item in Gio.AppInfo.get_all():
            if AppSource.should_show(item, desktop_type, use_filter):
                continue
            if not item.supports_uris() and not item.supports_files():
                continue
            yield AppLeaf(item)

    def should_sort_lexically(self):
        return False

    def get_description(self):
        return None

    def get_icon_name(self):
        return "applications-office"

    def provides(self):
        yield AppLeaf

class ResetAssociations (Action):
    rank_adjust = -10
    def __init__(self):
        super().__init__(_("Reset Associations"))

    def activate(self, leaf):
        content_type = leaf.get_content_type()
        Gio.AppInfo.reset_type_associations(content_type)

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Reset program associations for files of this type.")

